# -*- coding: utf-8 -*-

# Copyright 2019-2025 Mike Fährmann
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.

"""Extractors for https://nozomi.la/"""

from .common import Extractor, Message
from .. import text, dt


def decode_nozomi(n):
    for i in range(0, len(n), 4):
        yield (n[i] << 24) + (n[i+1] << 16) + (n[i+2] << 8) + n[i+3]


class NozomiExtractor(Extractor):
    """Base class for nozomi extractors"""
    category = "nozomi"
    root = "https://nozomi.la"
    domain = "gold-usergeneratedcontent.net"
    filename_fmt = "{postid} {dataid}.{extension}"
    archive_fmt = "{dataid}"

    def _init(self):
        self.session.headers["Origin"] = self.root

    def items(self):
        data = self.metadata()

        for post_id in map(str, self.posts()):
            url = (f"https://j.{self.domain}/post"
                   f"/{post_id[-1]}/{post_id[-3:-1]}/{post_id}.json")
            response = self.request(url, fatal=False)

            if response.status_code >= 400:
                self.log.warning(
                    "Skipping post %s ('%s %s')",
                    post_id, response.status_code, response.reason)
                continue

            post = response.json()
            post["tags"] = self._list(post.get("general"))
            post["artist"] = self._list(post.get("artist"))
            post["copyright"] = self._list(post.get("copyright"))
            post["character"] = self._list(post.get("character"))

            try:
                post["date"] = dt.parse_iso(post["date"] + ":00")
            except Exception:
                post["date"] = dt.NONE

            post.update(data)

            images = post["imageurls"]
            for key in ("general", "imageurl", "imageurls"):
                if key in post:
                    del post[key]

            yield Message.Directory, "", post
            for post["num"], image in enumerate(images, 1):
                post["filename"] = post["dataid"] = did = image["dataid"]
                post["is_video"] = video = \
                    True if image.get("is_video") else False

                ext = image["type"]
                if video:
                    subdomain = "v"
                elif ext == "gif":
                    subdomain = "g"
                else:
                    subdomain = "w"
                    ext = "webp"

                post["extension"] = ext
                post["url"] = url = (f"https://{subdomain}.{self.domain}"
                                     f"/{did[-1]}/{did[-3:-1]}/{did}.{ext}")
                yield Message.Url, url, post

    def posts(self):
        url = "https://n.nozomi.la" + self.nozomi
        offset = (text.parse_int(self.pnum, 1) - 1) * 256

        while True:
            headers = {"Range": f"bytes={offset}-{offset + 255}"}
            response = self.request(url, headers=headers)
            yield from decode_nozomi(response.content)

            offset += 256
            cr = response.headers.get("Content-Range", "").rpartition("/")[2]
            if text.parse_int(cr, offset) <= offset:
                return

    def metadata(self):
        return {}

    def _list(self, src):
        return [x["tagname_display"] for x in src] if src else ()


class NozomiPostExtractor(NozomiExtractor):
    """Extractor for individual posts on nozomi.la"""
    subcategory = "post"
    pattern = r"(?:https?://)?nozomi\.la/post/(\d+)"
    example = "https://nozomi.la/post/12345.html"

    def __init__(self, match):
        NozomiExtractor.__init__(self, match)
        self.post_id = match[1]

    def posts(self):
        return (self.post_id,)


class NozomiIndexExtractor(NozomiExtractor):
    """Extractor for the nozomi.la index"""
    subcategory = "index"
    pattern = (r"(?:https?://)?nozomi\.la/"
               r"(?:(index(?:-Popular)?)-(\d+)\.html)?(?:$|#|\?)")
    example = "https://nozomi.la/index-1.html"

    def __init__(self, match):
        NozomiExtractor.__init__(self, match)
        index, self.pnum = match.groups()
        self.nozomi = f"/{index or 'index'}.nozomi"


class NozomiTagExtractor(NozomiExtractor):
    """Extractor for posts from tag searches on nozomi.la"""
    subcategory = "tag"
    directory_fmt = ("{category}", "{search_tags}")
    archive_fmt = "t_{search_tags}_{dataid}"
    pattern = r"(?:https?://)?nozomi\.la/tag/([^/?#]+)-(\d+)\."
    example = "https://nozomi.la/tag/TAG-1.html"

    def __init__(self, match):
        NozomiExtractor.__init__(self, match)
        tags, self.pnum = match.groups()
        self.tags = text.unquote(tags)
        self.nozomi = f"/nozomi/{self.tags}.nozomi"

    def metadata(self):
        return {"search_tags": self.tags}


class NozomiSearchExtractor(NozomiExtractor):
    """Extractor for search results on nozomi.la"""
    subcategory = "search"
    directory_fmt = ("{category}", "{search_tags:J }")
    archive_fmt = "t_{search_tags}_{dataid}"
    pattern = r"(?:https?://)?nozomi\.la/search\.html\?q=([^&#]+)"
    example = "https://nozomi.la/search.html?q=QUERY"

    def __init__(self, match):
        NozomiExtractor.__init__(self, match)
        self.tags = text.unquote(match[1]).split()

    def metadata(self):
        return {"search_tags": self.tags}

    def posts(self):
        result = None
        positive = []
        negative = []

        def nozomi(path):
            url = f"https://j.{self.domain}/{path}.nozomi"
            return decode_nozomi(self.request(url).content)

        for tag in self.tags:
            (negative if tag[0] == "-" else positive).append(
                text.quote(tag.replace("/", "")))

        for tag in positive:
            ids = nozomi("nozomi/" + tag)
            if result is None:
                result = set(ids)
            else:
                result.intersection_update(ids)

        if result is None:
            result = set(nozomi("index"))
        for tag in negative:
            result.difference_update(nozomi("nozomi/" + tag[1:]))

        return sorted(result, reverse=True) if result else ()



# -*- coding: utf-8 -*-

# Copyright 2025 Mike Fährmann
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.

"""Extractors for https://imgpile.com/"""

from .common import Extractor, Message
from .. import text

BASE_PATTERN = r"(?:https?://)?(?:www\.)?imgpile\.com"


class ImgpileExtractor(Extractor):
    """Base class for imgpile extractors"""
    category = "imgpile"
    root = "https://imgpile.com"
    directory_fmt = ("{category}", "{post[author]}",
                     "{post[title]} ({post[id_slug]})")
    archive_fmt = "{post[id_slug]}_{id}"


class ImgpilePostExtractor(ImgpileExtractor):
    subcategory = "post"
    pattern = BASE_PATTERN + r"/p/(\w+)"
    example = "https://imgpile.com/p/AbCdEfG"

    def items(self):
        post_id = self.groups[0]
        url = f"{self.root}/p/{post_id}"
        page = self.request(url).text
        extr = text.extract_from(page)

        post = {
            "id_slug": post_id,
            "title"  : text.unescape(extr("<title>", " - imgpile<")),
            "id"     : text.parse_int(extr('data-post-id="', '"')),
            "author" : extr('/u/', '"'),
            "score"  : text.parse_int(text.remove_html(extr(
                'class="post-score">', "</"))),
            "views"  : text.parse_int(extr(
                'class="meta-value">', "<").replace(",", "")),
            "tags"   : text.split_html(extr(
                " <!-- Tags -->", '<!-- "')),
        }

        files = self._extract_files(extr)
        data = {"post": post}
        data["count"] = post["count"] = len(files)

        yield Message.Directory, "", data
        for data["num"], file in enumerate(files, 1):
            data.update(file)
            url = file["url"]
            yield Message.Url, url, text.nameext_from_url(url, data)

    def _extract_files(self, extr):
        files = []

        while True:
            media = extr('lass="post-media', '</div>')
            if not media:
                break
            files.append({
                "id_slug": text.extr(media, 'data-id="', '"'),
                "id" : text.parse_int(text.extr(
                    media, 'data-media-id="', '"')),
                "url": "http" + text.extr(media, '<a href="http', '"'),
            })
        return files


class ImgpileUserExtractor(ImgpileExtractor):
    subcategory = "user"
    pattern = BASE_PATTERN + r"/u/([^/?#]+)"
    example = "https://imgpile.com/u/USER"

    def items(self):
        url = self.root + "/api/v1/posts"
        params = {
            "limit"     : "100",
            "sort"      : "latest",
            "period"    : "all",
            "visibility": "public",
            "username"  : self.groups[0],
        }
        headers = {
            "Accept"        : "application/json",
            #  "Referer"       : "https://imgpile.com/u/USER",
            "Content-Type"  : "application/json",
            #  "X-CSRF-TOKEN": "",
            "Sec-Fetch-Dest": "empty",
            "Sec-Fetch-Mode": "cors",
            "Sec-Fetch-Site": "same-origin",
        }

        base = self.root + "/p/"
        while True:
            data = self.request_json(url, params=params, headers=headers)

            if params is not None:
                params = None
                self.kwdict["total"] = data["meta"]["total"]

            for item in data["data"]:
                item["_extractor"] = ImgpilePostExtractor
                url = base + item["slug"]
                yield Message.Queue, url, item

            url = data["links"].get("next")
            if not url:
                return



# -*- coding: utf-8 -*-

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.

"""Extractors for https://nudostar.tv/"""

from .common import GalleryExtractor, Message
from .. import text

BASE_PATTERN = r"(?:https?://)?(?:[a-z]{2}.)?nudostar\.tv"


class NudostarExtractor(GalleryExtractor):
    """Base class for NudoStar extractors"""
    category = "nudostar"
    root = "https://nudostar.tv"


class NudostarModelExtractor(NudostarExtractor):
    """Extractor for NudoStar models"""
    subcategory = "model"
    pattern = BASE_PATTERN + r"(/models/([^/?#]+)/?)$"
    example = "https://nudostar.tv/models/MODEL/"

    def metadata(self, page):
        names = text.extr(page, "<title>", "<").rpartition(
            " Nude ")[0].split(" / ")
        slug = self.groups[1]

        return {
            "gallery_id" : slug,
            "model_slug" : slug,
            "model_names": names,
            "model"      : names[0],
            "title"      : "",
        }

    def images(self, page):
        path = text.extr(page, '" src="https://nudostar.tv', '"')
        path, cnt, end = path.rsplit("_", 2)

        base = f"{self.root}{path}_"
        ext = "." + end.rpartition(".")[2]

        return [
            (f"{base}{i:04}{ext}", None)
            for i in range(1, int(cnt)+1)
        ]


class NudostarImageExtractor(NudostarExtractor):
    """Extractor for NudoStar images"""
    subcategory = "image"
    pattern = BASE_PATTERN + r"(/models/([^/?#]+)/(\d+)/)"
    example = "https://nudostar.tv/models/MODEL/123/"

    def items(self):
        page = self.request(self.page_url, notfound=self.subcategory).text

        img_url = text.extract(
            page, 'src="', '"', page.index('class="headline"'))[0]

        data = NudostarModelExtractor.metadata(self, page)
        data = text.nameext_from_url(img_url, data)
        data["num"] = text.parse_int(self.groups[2])
        data["url"] = img_url

        yield Message.Directory, "", data
        yield Message.Url, img_url, data



# -*- coding: utf-8 -*-

# Copyright 2023-2025 Mike Fährmann
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.

"""Extractors for https://pixeldrain.com/"""

from .common import Extractor, Message
from .. import text, util

BASE_PATTERN = r"(?:https?://)?pixeldrain\.com"


class PixeldrainExtractor(Extractor):
    """Base class for pixeldrain extractors"""
    category = "pixeldrain"
    root = "https://pixeldrain.com"
    archive_fmt = "{id}"

    def _init(self):
        if api_key := self.config("api-key"):
            self.session.auth = util.HTTPBasicAuth("", api_key)

    def _available(self, file):
        if not file.get("availability"):
            return True

        self.status |= 1
        self.log.debug(file["availability"])
        self.log.warning(
            "%s: '%s'", file["name"], file["availability_message"])
        return False


class PixeldrainFileExtractor(PixeldrainExtractor):
    """Extractor for pixeldrain files"""
    subcategory = "file"
    filename_fmt = "{filename[:230]} ({id}).{extension}"
    pattern = BASE_PATTERN + r"/(?:u|api/file)/(\w+)"
    example = "https://pixeldrain.com/u/abcdefgh"

    def __init__(self, match):
        Extractor.__init__(self, match)
        self.file_id = match[1]

    def items(self):
        url = f"{self.root}/api/file/{self.file_id}"
        file = self.request_json(url + "/info")

        file["url"] = url + "?download"
        file["date"] = self.parse_datetime_iso(file["date_upload"])

        text.nameext_from_name(file["name"], file)
        yield Message.Directory, "", file
        if file.get("availability"):
            self.status |= 1
            self.log.debug(file["availability"])
            self.log.error("'%s'", file["availability_message"])
        else:
            yield Message.Url, file["url"], file


class PixeldrainAlbumExtractor(PixeldrainExtractor):
    """Extractor for pixeldrain albums"""
    subcategory = "album"
    directory_fmt = ("{category}",
                     "{album[date]:%Y-%m-%d} {album[title]} ({album[id]})")
    filename_fmt = "{num:>03} {filename[:230]} ({id}).{extension}"
    pattern = BASE_PATTERN + r"/(?:l|api/list)/(\w+)(?:#item=(\d+))?"
    example = "https://pixeldrain.com/l/abcdefgh"

    def __init__(self, match):
        Extractor.__init__(self, match)
        self.album_id = match[1]
        self.file_index = match[2]

    def items(self):
        url = f"{self.root}/api/list/{self.album_id}"
        album = self.request_json(url)
        album["date"] = self.parse_datetime_iso(album["date_created"])

        if self.config("zip", False):
            self.directory_fmt = ("{category}",)
            self.filename_fmt = "{filename[:230]} ({id}).{extension}"
            del album["files"]
            album["count"] = 1
            url += "/zip"

            file = {
                "id"   : album["id"],
                "url"  : url,
                "num"  : 0,
                "count": 1,
                "name" : album["title"] + ".zip",
                "date" : album["date"],
                "album": album,
                "filename" : album["title"],
                "extension": "zip",
            }

            yield Message.Directory, "", file
            yield Message.Url, url, file
            return

        files = album.pop("files")
        album["count"] = album.pop("file_count")
        if self.file_index:
            idx = text.parse_int(self.file_index)
            try:
                files = (files[idx],)
            except LookupError:
                files = ()
        else:
            idx = 0

        yield Message.Directory, "", {"album": album}
        for num, file in enumerate(files, idx+1):
            if not self._available(file):
                continue
            file["album"] = album
            file["num"] = num
            file["url"] = url = f"{self.root}/api/file/{file['id']}?download"
            file["date"] = self.parse_datetime_iso(file["date_upload"])
            text.nameext_from_name(file["name"], file)
            yield Message.Url, url, file


class PixeldrainFolderExtractor(PixeldrainExtractor):
    """Extractor for pixeldrain filesystem files and directories"""
    subcategory = "folder"
    filename_fmt = "{filename[:230]}.{extension}"
    archive_fmt = "{path}_{num}"
    pattern = BASE_PATTERN + r"/(?:d|api/filesystem)/([^?]+)"
    example = "https://pixeldrain.com/d/abcdefgh"

    def metadata(self, data):
        return {
            "type"       : data["type"],
            "path"       : data["path"],
            "name"       : data["name"],
            "mime_type"  : data["file_type"],
            "size"       : data["file_size"],
            "hash_sha256": data["sha256_sum"],
            "date"       : self.parse_datetime_iso(data["created"]),
        }

    def items(self):
        recursive = self.config("recursive")

        url = f"{self.root}/api/filesystem/{self.groups[0]}"
        stat = self.request_json(url + "?stat")

        paths = stat["path"]
        path = paths[stat["base_index"]]
        if path["type"] == "dir":
            children = [
                child
                for child in stat["children"]
                if child["name"] != ".search_index.gz"
            ]
        else:
            children = (path,)

        folder = self.metadata(path)
        folder["id"] = paths[0]["id"]

        yield Message.Directory, "", folder

        num = 0
        for child in children:
            if child["type"] == "file":
                num += 1
                if not self._available(child):
                    continue
                url = f"{self.root}/api/filesystem{child['path']}?attach"
                share_url = f"{self.root}/d{child['path']}"
                data = self.metadata(child)
                data.update({
                    "id"       : folder["id"],
                    "num"      : num,
                    "url"      : url,
                    "share_url": share_url,
                })
                data["filename"], _, data["extension"] = \
                    child["name"].rpartition(".")
                yield Message.Url, url, data

            elif child["type"] == "dir":
                if recursive:
                    url = f"{self.root}/d{child['path']}"
                    child["_extractor"] = PixeldrainFolderExtractor
                    yield Message.Queue, url, child

            else:
                self.log.debug("'%s' is of unknown type (%s)",
                               child.get("name"), child["type"])



# -*- coding: utf-8 -*-

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.

"""Extractors for https://pornstars.tube/"""

from .common import GalleryExtractor
from .. import text


class PornstarstubeGalleryExtractor(GalleryExtractor):
    """Extractor for image galleries from pornstars.tube"""
    category = "pornstarstube"
    root = "https://pornstars.tube"
    pattern = (r"(?:https?://)?(?:www\.)?pornstars\.tube"
               r"/albums/(\d+)(?:/([\w-]+))?")
    example = "https://pornstars.tube/albums/12345/SLUG/"

    def __init__(self, match):
        url = f"{self.root}/albums/{match[1]}/{match[2] or 'a'}/"
        GalleryExtractor.__init__(self, match, url)

    def metadata(self, page):
        gid, slug = self.groups
        return {
            "gallery_id": text.parse_int(gid),
            "slug"      : slug or "",
            "title"     : text.unescape(text.extr(
                page, "<title>", " - PORNSTARS.TUBE</title>")),
            "description": text.unescape(text.extr(
                page, 'name="description" content="', '"')),
            "tags": text.extr(
                page, 'name="keywords" content="', '"').split(", "),
        }

    def images(self, page):
        album = text.extr(page, 'class="block-album"', "\n</div>")
        return [
            (url, None)
            for url in text.extract_iter(album, ' href="', '"')
        ]



# -*- coding: utf-8 -*-

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.

"""Extractors for https://myhentaigallery.com/"""

from .common import Extractor, GalleryExtractor, Message
from .. import text, exception

BASE_PATTERN = r"(?:https?://)?myhentaigallery\.com"


class MyhentaigalleryBase():
    category = "myhentaigallery"
    root = "https://myhentaigallery.com"


class MyhentaigalleryGalleryExtractor(MyhentaigalleryBase, GalleryExtractor):
    """Extractor for image galleries from myhentaigallery.com"""
    directory_fmt = ("{category}", "{gallery_id} {artist:?[/] /J, }{title}")
    pattern = BASE_PATTERN + r"/g(?:allery/(?:thumbnails|show))?/(\d+)"
    example = "https://myhentaigallery.com/g/12345"

    def __init__(self, match):
        self.gallery_id = match[1]
        url = f"{self.root}/g/{self.gallery_id}"
        GalleryExtractor.__init__(self, match, url)

    def _init(self):
        self.session.headers["Referer"] = self.page_url

    def metadata(self, page):
        extr = text.extract_from(page)
        split = text.split_html

        title = extr('<div class="comic-description">\n', '</h1>').lstrip()
        if title.startswith("<h1>"):
            title = title[4:]

        if not title:
            raise exception.NotFoundError("gallery")

        return {
            "title"     : text.unescape(title),
            "gallery_id": text.parse_int(self.gallery_id),
            "tags"      : split(extr("        Categories:", "</div>")),
            "artist"    : split(extr("        Artists:"   , "</div>")),
            "group"     : split(extr("        Groups:"    , "</div>")),
            "parodies"  : split(extr("        Parodies:"  , "</div>")),
        }

    def images(self, page):
        return [
            (text.unescape(text.extr(url, 'src="', '"')).replace(
                "/thumbnail/", "/original/"), None)
            for url in text.extract_iter(page, 'class="comic-thumb"', '</div>')
        ]


class MyhentaigalleryTagExtractor(MyhentaigalleryBase, Extractor):
    """Extractor for myhentaigallery tag searches"""
    subcategory = "tag"
    pattern = BASE_PATTERN + r"(/g/(artist|category|group|parody)/(\d+).*)"
    example = "https://myhentaigallery.com/g/category/123"

    def items(self):
        data = {"_extractor": MyhentaigalleryGalleryExtractor}
        for url in self.galleries():
            yield Message.Queue, url, data

    def galleries(self):
        root = self.root
        url = root + self.groups[0]

        while True:
            page = self.request(url).text

            for inner in text.extract_iter(
                    page, '<div class="comic-inner">', "<div"):
                yield root + text.extr(inner, 'href="', '"')

            try:
                pos = page.index(">Next<")
            except ValueError:
                return
            url = root + text.rextr(page, 'href="', '"', pos)



# -*- coding: utf-8 -*-

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.

"""Extractors for https://pictoa.com/"""

from .common import Extractor, Message
from .. import text

BASE_PATTERN = r"(?:https?://)?(?:[\w]+\.)?pictoa\.com(?:\.de)?"


class PictoaExtractor(Extractor):
    """Base class for pictoa extractors"""
    category = "pictoa"
    root = "https://pictoa.com"
    directory_fmt = ("{category}", "{album_id} {album_title}")
    filename_fmt = "{id}.{extension}"
    archive_fmt = "{id}"


class PictoaImageExtractor(PictoaExtractor):
    """Extractor for single images from pictoa.com"""
    subcategory = "image"
    pattern = BASE_PATTERN + r"/albums/(?:[\w-]+-)?(\d+)/(\d+)"
    example = "https://www.pictoa.com/albums/NAME-12345/12345.html"

    def items(self):
        album_id, image_id = self.groups

        url = f"{self.root}/albums/{album_id}/{image_id}.html"
        page = self.request(url).text
        album_title = text.extr(page, 'property="og:title" content="', '"')
        image_url = text.extr(page, 'property="og:image" content="', '"')

        data = {
            "album_id"   : album_id,
            "album_title": album_title.rpartition(" #")[0],
            "id"         : image_id,
            "url"        : image_url,
        }

        text.nameext_from_url(image_url, data)
        yield Message.Directory, "", data
        yield Message.Url, image_url, data


class PictoaAlbumExtractor(PictoaExtractor):
    """Extractor for image albums from pictoa.com"""
    subcategory = "album"
    pattern = BASE_PATTERN + r"/albums/(?:[\w-]+-)?(\d+).html"
    example = "https://www.pictoa.com/albums/NAME-12345.html"

    def items(self):
        album_id = self.groups[0]
        url = f"{self.root}/albums/{album_id}.html"
        page = self.request(url).text

        album_data = {
            "album_id"   : album_id,
            "album_title": text.extr(page, "<h1>", "<"),
            "tags"       : text.split_html(text.extr(
                page, '<ol class="related-categories', '</ol>'))[1:],
            "_extractor" : PictoaImageExtractor,
        }

        while True:
            container = text.extr(page, '<main>', '<span id="flag" >')
            for url in text.extract_iter(
                    container, '<a rel="nofollow" href="', '"'):
                yield Message.Queue, url, album_data

            url = text.extr(page, '<link rel="next" href="', '"')
            if not url:
                break
            page = self.request(url).text



# -*- coding: utf-8 -*-

# Copyright 2025 Mike Fährmann
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.

"""Extractors for https://thehentaiworld.com/"""

from .common import Extractor, Message
from .. import text, util
import collections

BASE_PATTERN = r"(?:https?://)?(?:www\.)?thehentaiworld\.com"


class ThehentaiworldExtractor(Extractor):
    """Base class for thehentaiworld extractors"""
    category = "thehentaiworld"
    root = "https://thehentaiworld.com"
    filename_fmt = "{title} ({id}{num:?-//}).{extension}"
    archive_fmt = "{id}_{num}"
    request_interval = (0.5, 1.5)

    def items(self):
        for url in self.posts():
            try:
                post = self._extract_post(url)
            except Exception as exc:
                self.status |= 1
                self.log.warning("Failed to extract post %s (%s: %s)",
                                 url, exc.__class__.__name__, exc)
                continue

            if "file_urls" in post:
                urls = post["file_urls"]
                post["count"] = len(urls)
                yield Message.Directory, "", post
                for post["num"], url in enumerate(urls, 1):
                    text.nameext_from_url(url, post)
                    yield Message.Url, url, post
            else:
                yield Message.Directory, "", post
                url = post["file_url"]
                text.nameext_from_url(url, post)
                yield Message.Url, url, post

    def _extract_post(self, url):
        extr = text.extract_from(self.request(url).text)

        post = {
            "num"     : 0,
            "count"   : 1,
            "title"   : text.unescape(extr("<title>", "<").strip()),
            "id"      : text.parse_int(extr(" postid-", " ")),
            "slug"    : extr(" post-", '"'),
            "tags"    : extr('id="tagsHead">', "</ul>"),
            "date"    : self.parse_datetime_iso(extr("<li>Posted: ", "<")),
        }

        if (c := url[27]) == "v":
            post["type"] = "video"
            post["width"] = post["height"] = 0
            post["votes"] = text.parse_int(extr("(<strong>", "</strong>"))
            post["score"] = text.parse_float(extr("<strong>", "<"))
            post["file_url"] = extr('<source src="', '"')
        else:
            post["type"] = ("animated" if c == "g" else
                            "3d cgi" if c == "3" else
                            "image")
            post["width"] = text.parse_int(extr("<li>Size: ", " "))
            post["height"] = text.parse_int(extr("x ", "<"))
            post["file_url"] = extr('a href="', '"')
            post["votes"] = text.parse_int(extr("(<strong>", "</strong>"))
            post["score"] = text.parse_float(extr("<strong>", "<"))

            if doujin := extr('<a id="prev-page"', "</div></div><"):
                repl = text.re(r"-220x\d+\.").sub
                post["file_urls"] = [
                    repl(".", url)
                    for url in text.extract_iter(
                        doujin, 'class="border" src="', '"')
                ]

        tags = collections.defaultdict(list)
        pattern = text.re(r'<li><a class="([^"]*)" href="[^"]*">([^<]+)')
        for tag_type, tag_name in pattern.findall(post["tags"]):
            tags[tag_type].append(tag_name)
        post["tags"] = tags_list = []
        for key, value in tags.items():
            tags_list.extend(value)
            post["tags_" + key if key else "tags_general"] = value

        return post

    def _pagination(self, endpoint):
        base = self.root + endpoint
        pnum = self.page_start

        while True:
            url = base if pnum < 2 else f"{base}page/{pnum}/"
            page = self.request(url).text

            yield from text.extract_iter(text.extr(
                page, 'id="thumbContainer"', "<script"), ' href="', '"')

            if 'class="next"' not in page:
                return
            pnum += 1


class ThehentaiworldTagExtractor(ThehentaiworldExtractor):
    subcategory = "tag"
    per_page = 24
    page_start = 1
    post_start = 0
    directory_fmt = ("{category}", "{search_tags}")
    pattern = BASE_PATTERN + r"/tag/([^/?#]+)"
    example = "https://thehentaiworld.com/tag/TAG/"

    def posts(self):
        self.kwdict["search_tags"] = tag = self.groups[0]
        return util.advance(self._pagination(f"/tag/{tag}/"), self.post_start)

    def skip(self, num):
        pages, posts = divmod(num, self.per_page)
        self.page_start += pages
        self.post_start += posts
        return num


class ThehentaiworldPostExtractor(ThehentaiworldExtractor):
    subcategory = "post"
    pattern = (BASE_PATTERN +
               r"(/(?:video|(?:[\w-]+-)?hentai-image)s/([^/?#]+))")
    example = "https://thehentaiworld.com/hentai-images/SLUG/"

    def posts(self):
        return (f"{self.root}{self.groups[0]}/",)



# -*- coding: utf-8 -*-

# Copyright 2024-2026 Mike Fährmann
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.

"""Extractors for https://turbo.cr/"""

from .lolisafe import LolisafeAlbumExtractor
from .. import text

BASE_PATTERN = (r"(?:https?://)?(?:"
                r"(?:www\.)?turbo(?:vid)?\.cr|"
                r"saint\d*\.(?:su|pk|cr|to))")


class TurboAlbumExtractor(LolisafeAlbumExtractor):
    """Extractor for turbo.cr albums"""
    category = "turbo"
    root = "https://turbo.cr"
    pattern = BASE_PATTERN + r"/a/([^/?#]+)"
    example = "https://turbo.cr/a/ID"

    def fetch_album(self, album_id):
        url = f"{self.root}/a/{album_id}"
        extr = text.extract_from(self.request(url).text)
        title = extr("<h1 ", "<")
        descr = extr("<p ", "<")
        tbody = extr('id="fileTbody"', '</tbody>')
        headers = {"Referer": url}

        return self._extract_files(tbody, headers), {
            "album_id"     : album_id,
            "album_name"   : text.unescape(title[title.find(">")+1:]),
            "description"  : text.unescape(descr[descr.find(">")+1:]),
            "album_size"   : sum(map(text.parse_int, text.extract_iter(
                tbody, 'data-size="', '"'))),
            "count"        : tbody.count("data-id="),
            "_http_headers": headers,
        }

    def _extract_files(self, body, headers):
        for file in text.extract_iter(body, "<tr", "</tr>"):
            data_id = text.extr(file, 'data-id="', '"')
            url = f"{self.root}/api/sign?v={data_id}"
            data = self.request_json(url, headers=headers)
            name = data.get("original_filename") or data.get("filename")
            yield text.nameext_from_name(name, {
                "id"  : data_id,
                "file": data.get("url"),
                "size": text.parse_int(text.extr(file, 'data-size="', '"')),
                "_http_headers": headers,
            })


class TurboMediaExtractor(TurboAlbumExtractor):
    """Extractor for turbo.cr media links"""
    subcategory = "media"
    directory_fmt = ("{category}",)
    pattern = BASE_PATTERN + r"/(?:embe)?[dv]/([^/?#]+)"
    example = "https://turbo.cr/embed/ID"

    def fetch_album(self, album_id):
        try:
            return (self._extract_file(album_id),), {
                "album_id"   : "",
                "album_name" : "",
                "album_size" : -1,
                "description": "",
                "count"      : 1,
            }
        except Exception as exc:
            self.log.error("%s: %s", exc.__class__.__name__, exc)
            return (), {}

    def _extract_file(self, data_id):
        url = f"{self.root}/d/{data_id}"
        headers = {"Referer": url}
        page = self.request(url).text
        size = text.extr(page, 'id="fileSizeBytes">', '<')
        date = text.extract(page, "<span>", "<", page.find("File ID:"))[0]

        url = f"{self.root}/api/sign?v={data_id}"
        data = self.request_json(url, headers=headers)
        name = data.get("original_filename") or data.get("filename")
        return text.nameext_from_name(name, {
            "id"  : data_id,
            "file": data.get("url"),
            "size": int(text.parse_float(size.replace("&#43;", "+"))),
            "date": self.parse_datetime_iso(date),
            "_http_headers": headers,
        })



# -*- coding: utf-8 -*-

# Copyright 2025-2026 Mike Fährmann
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.

"""Extractors for https://www.bellazon.com/"""

from .common import Extractor, Message
from .. import text, exception

BASE_PATTERN = r"(?:https?://)?(?:www\.)?bellazon\.com/main"


class BellazonExtractor(Extractor):
    """Base class for bellazon extractors"""
    category = "bellazon"
    root = "https://www.bellazon.com/main"
    directory_fmt = ("{category}", "{thread[section]}",
                     "{thread[title]} ({thread[id]})")
    filename_fmt = "{post[id]}_{num:>02}_{id}_{filename}.{extension}"
    archive_fmt = "{post[id]}/{id}_{filename}"

    def items(self):
        native = (self.root + "/", self.root[6:] + "/")
        quotes = self.config("quoted", False)
        extract_urls = text.re(
            r'(?s)<('
            r'(?:video .*?<source [^>]*?src|a [^>]*?href)="([^"]+).*?</a>'
            r'|img [^>]*?src="([^"]+)"[^>]*>'
            r')'
        ).findall

        for post in self.posts():
            if quotes:
                urls = extract_urls(post["content"])
            else:
                urls = extract_urls(self._remove_quotes(post["content"]))

            data = {"post": post}
            post["count"] = data["count"] = len(urls)

            yield Message.Directory, "", data
            data["num"] = data["num_internal"] = data["num_external"] = 0
            for info, url, url_img in urls:
                if url_img:
                    url = text.unescape(
                        text.extr(info, 'data-full-image="', '"') or url_img)
                else:
                    url = text.unescape(url)

                if url.startswith(native):
                    if (
                        "/uploads/emoticons/" in url or
                        "/public/style_" in url or
                        "/profile/" in url or
                        "/topic/" in url
                    ):
                        continue
                    data["num"] += 1
                    data["num_internal"] += 1
                    if not (alt := text.extr(info, ' alt="', '"')) or (
                            alt.startswith("post-") and "_thumb." in alt):
                        dc = text.nameext_from_url(url, data.copy())
                    else:
                        dc = data.copy()
                        dc["name"] = name = text.unescape(alt)
                        dc["filename"] = name.partition(".")[0]

                    dc["id"] = text.extr(info, 'data-fileid="', '"')
                    if ext := text.extr(info, 'data-fileext="', '"'):
                        dc["extension"] = ext
                    elif "/core/interface/file/attachment.php" in url:
                        if not dc["id"]:
                            dc["id"] = \
                                url.rpartition("?id=")[2].partition("&")[0]
                        if name := text.extr(info, ">", "<").strip():
                            dc["name"] = name = text.unescape(name)
                            text.nameext_from_name(name, dc)
                    else:
                        dc["extension"] = text.ext_from_url(url)

                    if url[0] == "/":
                        url = "https:" + url
                    yield Message.Url, url, dc

                else:
                    data["num"] += 1
                    data["num_external"] += 1
                    yield Message.Queue, url, data

    def _pagination(self, base, pnum=None):
        base = self.root + base

        if pnum is None:
            url = base + "/"
            pnum = 1
        else:
            url = f"{base}/page/{pnum}/"
            pnum = None

        while True:
            page = self.request(url).text

            yield page

            if pnum is None or ' rel="next" ' not in page or text.extr(
                    page, " rel=\"next\" data-page='", "'") == str(pnum):
                return
            pnum += 1
            url = f"{base}/page/{pnum}/"

    def _pagination_reverse(self, base, pnum=None):
        base = self.root + base

        url = f"{base}/page/{'9999' if pnum is None else pnum}/"
        with self.request(url) as response:
            parts = response.url.rsplit("/", 3)
            pnum = text.parse_int(parts[2]) if parts[1] == "page" else 1
            page = response.text

        while True:
            yield page

            pnum -= 1
            if pnum > 1:
                url = f"{base}/page/{pnum}/"
            elif pnum == 1:
                url = base + "/"
            else:
                return

            page = self.request(url).text

    def _parse_thread(self, page):
        schema = self._extract_jsonld(page)
        author = schema["author"]
        stats = schema["interactionStatistic"]
        url_t = schema["url"]
        url_a = author.get("url") or ""

        path = text.split_html(text.extr(
            page, '<nav class="ipsBreadcrumb', "</nav>"))[2:-1]

        thread = {
            "url"  : url_t,
            "path" : path,
            "title": schema["headline"],
            "views": stats[0]["userInteractionCount"],
            "posts": stats[1]["userInteractionCount"],
            "date" : self.parse_datetime_iso(schema["datePublished"]),
            "date_updated": self.parse_datetime_iso(schema["dateModified"]),
            "description" : text.unescape(schema["text"]).strip(),
            "section"     : path[-2],
            "author"      : author["name"],
            "author_url"  : url_a,
        }

        thread["id"], _, slug = \
            url_t.rsplit("/", 2)[1].partition("-")
        thread["slug"] = text.unquote(slug)

        if url_a:
            thread["author_id"], _, thread["author_slug"] = \
                url_a.rsplit("/", 2)[1].partition("-")
        else:
            thread["author_id"] = thread["author_slug"] = ""

        return thread

    def _parse_post(self, html):
        extr = text.extract_from(html)

        post = {
            "id": extr('id="elComment_', '"'),
            "author_url": extr(" href='", "'"),
            "date": self.parse_datetime_iso(extr("datetime='", "'")),
            "content": extr("<!-- Post content -->", '<menu data-ips-hook='),
        }

        beg = post["content"].find(">")
        sig = post["content"].rfind('<div data-role="memberSignature"')
        end = post["content"].rfind("\n\t\t</div>", 0, sig+1 or None)
        post["content"] = post["content"][beg+1:end+1].strip()

        if url_a := post["author_url"]:
            post["author_id"], _, post["author_slug"] = \
                url_a.rsplit("/", 2)[1].partition("-")
        else:
            post["author_id"] = post["author_slug"] = ""

        return post

    def _remove_quotes(self, content):
        while "<blockquote" in content:
            beg = content.index("<blockquote")
            end = content.index("</blockquote", beg)
            for _ in range(content.count("<blockquote", beg+11, end)):
                end = content.index("</blockquote", end+13)
            content = content[:beg] + content[end+13:]
        return content


class BellazonPostExtractor(BellazonExtractor):
    subcategory = "post"
    pattern = (BASE_PATTERN + r"(/topic/\d+-[^/?#]+(?:/page/\d+)?)"
               r"/?#(?:findC|c)omment-(\d+)")
    example = "https://www.bellazon.com/main/topic/123-SLUG/#findComment-12345"

    def posts(self):
        path, post_id = self.groups
        page = self.request(self.root + path).text

        pos = page.find('id="elComment_' + post_id)
        if pos < 0:
            raise exception.NotFoundError("post")
        html = text.extract(page, "<article ", "</article>", pos-100)[0]

        self.kwdict["thread"] = self._parse_thread(page)
        return (self._parse_post(html),)


class BellazonThreadExtractor(BellazonExtractor):
    subcategory = "thread"
    pattern = BASE_PATTERN + r"(/topic/\d+-[^/?#]+)(?:/page/(\d+))?"
    example = "https://www.bellazon.com/main/topic/123-SLUG/"

    def posts(self):
        if (order := self.config("order-posts")) and \
                order[0] not in ("d", "r"):
            pages = self._pagination(*self.groups)
            reverse = False
        else:
            pages = self._pagination_reverse(*self.groups)
            reverse = True

        for page in pages:
            if "thread" not in self.kwdict:
                self.kwdict["thread"] = self._parse_thread(page)
            posts = text.extract_iter(page, "<article ", "</article>")
            if reverse:
                posts = list(posts)
                posts.reverse()
            for html in posts:
                yield self._parse_post(html)


class BellazonForumExtractor(BellazonExtractor):
    subcategory = "forum"
    pattern = BASE_PATTERN + r"(/forum/\d+-[^/?#]+)(?:/page/(\d+))?"
    example = "https://www.bellazon.com/main/forum/123-SLUG/"

    def items(self):
        data = {"_extractor": BellazonThreadExtractor}
        for page in self._pagination(*self.groups):
            for row in text.extract_iter(
                    page, '<li data-ips-hook="topicRow"', "</"):
                yield Message.Queue, text.extr(row, 'href="', '"'), data



# -*- coding: utf-8 -*-

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.

"""Extractors for https://chzzk.naver.com/"""

from .common import Extractor, Message
from .. import text, util


class NaverChzzkExtractor(Extractor):
    """Base class for chzzk.naver.com extractors"""
    category = "naver-chzzk"
    filename_fmt = "{uid}_{id}_{num}.{extension}"
    directory_fmt = ("{category}", "{user[userNickname]}")
    archive_fmt = "{uid}_{id}_{num}"

    def request_api(self, uid, id=None, params=None):
        return self.request_json(
            f"https://apis.naver.com/nng_main/nng_comment_api/v1/type"
            f"/CHANNEL_POST/id/{uid}/comments/{id or ''}",
            params=params)["content"]

    def items(self):
        for comment in self.comments():
            data = comment["comment"]
            files = data.pop("attaches") or ()
            data["id"] = data["commentId"]
            data["uid"] = data["objectId"]
            data["user"] = comment["user"]
            data["count"] = len(files)
            data["date"] = self.parse_datetime(
                data["createdDate"], "%Y%m%d%H%M%S")

            yield Message.Directory, "", data
            for data["num"], file in enumerate(files, 1):
                if extra := file.get("extraJson"):
                    file.update(util.json_loads(extra))
                file["date"] = self.parse_datetime_iso(
                    file["createdDate"])
                file["date_updated"] = self.parse_datetime_iso(
                    file["updatedDate"])
                data["file"] = file
                url = file["attachValue"]
                yield Message.Url, url, text.nameext_from_url(url, data)


class NaverChzzkCommentExtractor(NaverChzzkExtractor):
    """Extractor for individual comment from chzzk.naver.com"""
    subcategory = "comment"
    pattern = r"(?:https?://)?chzzk\.naver\.com/(\w+)/community/detail/(\d+)"
    example = "https://chzzk.naver.com/0123456789abcdef/community/detail/12345"

    def comments(self):
        uid, id = self.groups
        res = self.request_api(uid, id)
        return ({"comment": res["comment"], "user": res["user"]},)


class NaverChzzkCommunityExtractor(NaverChzzkExtractor):
    """Extractor for comments from chzzk.naver.com"""
    subcategory = "community"
    pattern = r"(?:https?://)?chzzk\.naver\.com/(\w+)/community"
    example = "https://chzzk.naver.com/0123456789abcdef/community"
    request_interval = (0.5, 1.5)

    def comments(self):
        uid = self.match[1]
        params = {
            "limit": 10,
            "offset": text.parse_int(self.config("offset")),
            "pagingType": "PAGE",
        }
        while True:
            comments = self.request_api(uid, params=params)["comments"]
            yield from comments["data"]
            if not comments["page"]["next"]:
                return
            params["offset"] += params["limit"]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2021-2026 Mike Fährmann
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.

import os
import sys
import time
import unittest
import datetime
import tempfile

sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
from gallery_dl import formatter, text, dt, util, config  # noqa E402

try:
    import jinja2
except ImportError:
    jinja2 = None


class TestFormatter(unittest.TestCase):

    def tearDown(self):
        config.clear()

    kwdict = {
        "a": "hElLo wOrLd",
        "b": "äöü",
        "j": "げんそうきょう",
        "d": {"a": "foo", "b": 0, "c": None},
        "i": 2,
        "l": ["a", "b", "c"],
        "L": [
            {"name": "John Doe"      , "age": 42, "email": "jd@example.org"},
            {"name": "Jane Smith"    , "age": 24, "email": None},
            {"name": "Max Mustermann", "age": False},
        ],
        "n": None,
        "s": " \n\r\tSPACE    ",
        "S": " \n\r\tS  P         A\tC\nE    ",
        "h": "<p>foo </p> &amp; bar <p> </p>",
        "H": """<p>
  <a href="http://www.example.com">Lorem ipsum dolor sit amet</a>.
  Duis aute irure <a href="http://blog.example.org/lorem?foo=bar">
  http://blog.example.org</a>.
</p>""",
        "u": "&#x27;&lt; / &gt;&#x27;",
        "t": 1262304000,
        "ds": "2010-01-01T01:00:00+01:00",
        "dt": datetime.datetime(2010, 1, 1),
        "dt_dst": datetime.datetime(2010, 6, 1),
        "i_str": "12345",
        "f_str": "12.45",
        "lang": "en",
        "name": "Name",
        "title1": "Title",
        "title2": "",
        "title3": None,
        "title4": 0,
    }

    def test_conversions(self):
        self._run_test("{a!l}", "hello world")
        self._run_test("{a!u}", "HELLO WORLD")
        self._run_test("{a!c}", "Hello world")
        self._run_test("{a!C}", "Hello World")
        self._run_test("{s!t}", "SPACE")
        self._run_test("{S!t}", "S  P         A\tC\nE")
        self._run_test("{a!U}", self.kwdict["a"])
        self._run_test("{u!U}", "'< / >'")
        self._run_test("{a!H}", self.kwdict["a"])
        self._run_test("{h!H}", "foo & bar")
        self._run_test("{u!H}", "'< / >'")
        self._run_test("{n!H}", "")
        self._run_test("{h!R}", [])
        self._run_test("{H!R}", ["http://www.example.com",
                                 "http://blog.example.org/lorem?foo=bar",
                                 "http://blog.example.org"])
        self._run_test("{a!s}", self.kwdict["a"])
        self._run_test("{a!r}", f"'{self.kwdict['a']}'")
        self._run_test("{a!a}", f"'{self.kwdict['a']}'")
        self._run_test("{b!a}", "'\\xe4\\xf6\\xfc'")
        self._run_test("{a!S}", self.kwdict["a"])
        self._run_test("{l!S}", "a, b, c")
        self._run_test("{n!S}", "")
        self._run_test("{t!d}", datetime.datetime(2010, 1, 1))
        self._run_test("{t!d:%Y-%m-%d}", "2010-01-01")
        self._run_test("{t!D}" , datetime.datetime(2010, 1, 1))
        self._run_test("{ds!D}", datetime.datetime(2010, 1, 1))
        self._run_test("{dt!D}", datetime.datetime(2010, 1, 1))
        self._run_test("{t!D:%Y-%m-%d}", "2010-01-01")
        self._run_test("{dt!T}", "1262304000")
        self._run_test("{l!j}", '["a","b","c"]')
        self._run_test("{dt!j}", '"2010-01-01 00:00:00"')
        self._run_test("{a!g}", "hello-world")
        self._run_test("{lang!L}", "English")
        self._run_test("{'fr'!L}", "French")
        self._run_test("{a!L}", None)
        self._run_test("{a!n}", 11)
        self._run_test("{l!n}", 3)
        self._run_test("{d!n}", 3)
        self._run_test("{s!W}", "SPACE")
        self._run_test("{S!W}", "S P A C E")
        self._run_test("{i_str!i}", 12345)
        self._run_test("{i_str!f}", 12345.0)
        self._run_test("{f_str!f}", 12.45)

        with self.assertRaises(KeyError):
            self._run_test("{a!q}", "hello world")

    def test_optional(self):
        self._run_test("{name}{title1}", "NameTitle")
        self._run_test("{name}{title1:?//}", "NameTitle")
        self._run_test("{name}{title1:? **/''/}", "Name **Title''")

        self._run_test("{name}{title2}", "Name")
        self._run_test("{name}{title2:?//}", "Name")
        self._run_test("{name}{title2:? **/''/}", "Name")

        self._run_test("{name}{title3}", "NameNone")
        self._run_test("{name}{title3:?//}", "Name")
        self._run_test("{name}{title3:? **/''/}", "Name")

        self._run_test("{name}{title4}", "Name0")
        self._run_test("{name}{title4:?//}", "Name")
        self._run_test("{name}{title4:? **/''/}", "Name")

    def test_missing(self):
        replacement = "None"

        self._run_test("{missing}", replacement)
        self._run_test("{missing.attr}", replacement)
        self._run_test("{missing[key]}", replacement)
        self._run_test("{missing:?a//}", "")

        self._run_test("{name[missing]}", replacement)
        self._run_test("{name[missing].attr}", replacement)
        self._run_test("{name[missing][key]}", replacement)
        self._run_test("{name[missing]:?a//}", "")

    def test_missing_custom_default(self):
        replacement = default = "foobar"
        self._run_test("{missing}"     , replacement, default)
        self._run_test("{missing.attr}", replacement, default)
        self._run_test("{missing[key]}", replacement, default)
        self._run_test("{missing:?a//}", f"a{default}", default)

    def test_fmt_func(self):
        self._run_test("{t}" , self.kwdict["t"] , None, int)
        self._run_test("{t}" , self.kwdict["t"] , None, util.identity)
        self._run_test("{dt}", self.kwdict["dt"], None, util.identity)
        self._run_test("{ds}", self.kwdict["dt"], None, dt.parse_iso)
        self._run_test("{ds:D%Y-%m-%dT%H:%M:%S%z}", self.kwdict["dt"],
                       None, util.identity)

    def test_fmt_func_multi(self):
        self._run_test("foo {t}" , f"foo {self.kwdict['t']}",
                       None, int)
        self._run_test("bar {ds}", f"bar {self.kwdict['dt']}",
                       None, dt.parse_iso)
        self._run_test("foo {ds:D%Y-%m-%dT%H:%M:%S%z} bar",
                       f"foo {self.kwdict['dt']} bar",
                       None, util.identity)

    def test_alternative(self):
        self._run_test("{a|z}"    , "hElLo wOrLd")
        self._run_test("{z|a}"    , "hElLo wOrLd")
        self._run_test("{z|y|a}"  , "hElLo wOrLd")
        self._run_test("{z|y|x|a}", "hElLo wOrLd")
        self._run_test("{z|n|a|y}", "hElLo wOrLd")

        self._run_test("{z|a!C}"      , "Hello World")
        self._run_test("{z|a:Rh/C/}"  , "CElLo wOrLd")
        self._run_test("{z|a!C:RH/C/}", "Cello World")
        self._run_test("{z|y|x:?</>/}", "")

        self._run_test("{d[c]|d[b]|d[a]}", "foo")
        self._run_test("{d[a]|d[b]|d[c]}", "foo")
        self._run_test("{d[z]|d[y]|d[x]}", "None")

    def test_indexing(self):
        self._run_test("{l[0]}" , "a")
        self._run_test("{a[6]}" , "w")

    def test_indexing_negative(self):
        self._run_test("{l[-1]}" , "c")
        self._run_test("{a[-7]}" , "o")
        self._run_test("{a[-0]}" , "h")  # same as a[0]

    def test_dict_access(self):
        self._run_test("{d[a]}"  , "foo")
        self._run_test("{d['a']}", "foo")
        self._run_test('{d["a"]}', "foo")

    def test_dot_index(self):
        self._run_test("{l.1}"  , "b")
        self._run_test("{a.6}"  , "w")
        self._run_test("{a.99}" , "None")
        self._run_test("{l.-1}" , "c")
        self._run_test("{a.-7}" , "o")
        self._run_test("{a.-0}" , "h")  # same as a[0]
        self._run_test("{a.-99}", "None")

    def test_dot_access_dict(self):
        self._run_test("{d.a}", "foo")
        self._run_test("{d.d}", "None")
        self._run_test("{a.d}", "None")
        self._run_test("{L.0.age}", "42")
        self._run_test("{L.-1.name.2}", "x")
        self._run_test("{L.1:I}", self.kwdict["L"][1])

    def test_dot_access_attr(self):
        self._run_test("{t.real}", "1262304000")
        self._run_test("{dt.year}", "2010")

    def test_slice_str(self):
        v = self.kwdict["a"]
        self._run_test("{a[1:10]}"  , v[1:10])
        self._run_test("{a[-10:-1]}", v[-10:-1])
        self._run_test("{a[5:]}" , v[5:])
        self._run_test("{a[50:]}", v[50:])
        self._run_test("{a[:5]}" , v[:5])
        self._run_test("{a[:50]}", v[:50])
        self._run_test("{a[:]}"  , v)
        self._run_test("{a[1:10:2]}"  , v[1:10:2])
        self._run_test("{a[-10:-1:2]}", v[-10:-1:2])
        self._run_test("{a[5::2]}" , v[5::2])
        self._run_test("{a[50::2]}", v[50::2])
        self._run_test("{a[:5:2]}" , v[:5:2])
        self._run_test("{a[:50:2]}", v[:50:2])
        self._run_test("{a[::]}"   , v)

        self._run_test("{a:[1:10]}"  , v[1:10])
        self._run_test("{a:[-10:-1]}", v[-10:-1])
        self._run_test("{a:[5:]}" , v[5:])
        self._run_test("{a:[50:]}", v[50:])
        self._run_test("{a:[:5]}" , v[:5])
        self._run_test("{a:[:50]}", v[:50])
        self._run_test("{a:[:]}"  , v)
        self._run_test("{a:[1:10:2]}"  , v[1:10:2])
        self._run_test("{a:[-10:-1:2]}", v[-10:-1:2])
        self._run_test("{a:[5::2]}" , v[5::2])
        self._run_test("{a:[50::2]}", v[50::2])
        self._run_test("{a:[:5:2]}" , v[:5:2])
        self._run_test("{a:[:50:2]}", v[:50:2])
        self._run_test("{a:[::]}"   , v)

    def test_slice_bytes(self):
        v = self.kwdict["j"]
        self._run_test("{j[b1:10]}"  , v[1:3])
        self._run_test("{j[b-10:-1]}", v[-3:-1])
        self._run_test("{j[b5:]}"    , v[2:])
        self._run_test("{j[b50:]}"   , v[50:])
        self._run_test("{j[b:5]}"    , v[:1])
        self._run_test("{j[b:50]}"   , v[:50])
        self._run_test("{j[b:]}"     , v)
        self._run_test("{j[b::]}"    , v)

        self._run_test("{j:[b1:10]}"  , v[1:3])
        self._run_test("{j:[b-10:-1]}", v[-3:-1])
        self._run_test("{j:[b5:]}"    , v[2:])
        self._run_test("{j:[b50:]}"   , v[50:])
        self._run_test("{j:[b:5]}"    , v[:1])
        self._run_test("{j:[b:50]}"   , v[:50])
        self._run_test("{j:[b:]}"     , v)
        self._run_test("{j:[b::]}"    , v)

    def test_specifier_maxlen(self):
        v = self.kwdict["a"]
        self._run_test("{a:L5/foo/}" , "foo")
        self._run_test("{a:L50/foo/}", v)
        self._run_test("{a:L50/foo/>50}", " " * 39 + v)
        self._run_test("{a:L50/foo/>51}", "foo")
        self._run_test("{a:Lab/foo/}", "foo")

    def test_specifier_maxlen_bytes(self):
        v = self.kwdict["a"]
        self._run_test("{a:Lb5/foo/}" , "foo")
        self._run_test("{a:Lb50/foo/}", v)
        self._run_test("{a:Lb50/foo/>50}", " " * 39 + v)
        self._run_test("{a:Lb50/foo/>51}", "foo")
        self._run_test("{a:Lbab/foo/}", "foo")

        v = self.kwdict["j"]
        self._run_test("{j:Lb5/foo/}" , "foo")
        self._run_test("{j:Lb50/foo/}", v)
        self._run_test("{j:Lbab/foo/}", "foo")

    def test_specifier_join(self):
        self._run_test("{l:J}"       , "abc")
        self._run_test("{l:J,}"      , "a,b,c")
        self._run_test("{l:J,/}"     , "a,b,c")
        self._run_test("{l:J,/>20}"  , "               a,b,c")
        self._run_test("{l:J - }"    , "a - b - c")
        self._run_test("{l:J - /}"   , "a - b - c")
        self._run_test("{l:J - />20}", "           a - b - c")

        self._run_test("{a:J/}"      , self.kwdict["a"])
        self._run_test("{a:J, /}"    , self.kwdict["a"])

    def test_specifier_replace(self):
        self._run_test("{a:Rh/C/}"  , "CElLo wOrLd")
        self._run_test("{a!l:Rh/C/}", "Cello world")
        self._run_test("{a!u:Rh/C/}", "HELLO WORLD")

        self._run_test("{a!l:Rl/_/}", "he__o wor_d")
        self._run_test("{a!l:Rl//}" , "heo word")
        self._run_test("{name:Rame/othing/}", "Nothing")

    def test_specifier_datetime(self):
        self._run_test("{ds:D%Y-%m-%dT%H:%M:%S%z}", "2010-01-01 00:00:00")
        self._run_test("{ds:D%Y}", "[Invalid DateTime]")
        self._run_test("{l2:D%Y}", "[Invalid DateTime]")

    def test_specifier_offset(self):
        self._run_test("{dt:O 01:00}", "2010-01-01 01:00:00")
        self._run_test("{dt:O+02:00}", "2010-01-01 02:00:00")
        self._run_test("{dt:O-03:45}", "2009-12-31 20:15:00")

        self._run_test("{dt:O12}", "2010-01-01 12:00:00")
        self._run_test("{dt:O-24}", "2009-12-31 00:00:00")

        self._run_test("{ds:D%Y-%m-%dT%H:%M:%S%z/O1}", "2010-01-01 01:00:00")
        self._run_test("{t!d:O2}", "2010-01-01 02:00:00")

    def test_specifier_offset_local(self):
        ts = self.kwdict["dt"].replace(
            tzinfo=datetime.timezone.utc).timestamp()
        offset = time.localtime(ts).tm_gmtoff
        dt = self.kwdict["dt"] + datetime.timedelta(seconds=offset)
        self._run_test("{dt:O}", str(dt))
        self._run_test("{dt:Olocal}", str(dt))

        ts = self.kwdict["dt_dst"].replace(
            tzinfo=datetime.timezone.utc).timestamp()
        offset = time.localtime(ts).tm_gmtoff
        dt = self.kwdict["dt_dst"] + datetime.timedelta(seconds=offset)
        self._run_test("{dt_dst:O}", str(dt))
        self._run_test("{dt_dst:Olocal}", str(dt))

    def test_specifier_sort(self):
        self._run_test("{l:S}" , "['a', 'b', 'c']")
        self._run_test("{l:Sa}", "['a', 'b', 'c']")
        self._run_test("{l:Sd}", "['c', 'b', 'a']")
        self._run_test("{l:Sr}", "['c', 'b', 'a']")

        self._run_test(
            "{a:S}", "[' ', 'E', 'L', 'L', 'O', 'd', 'h', 'l', 'o', 'r', 'w']")
        self._run_test(
            "{a:S-asc}",  # starts with 'S', contains 'a'
            "[' ', 'E', 'L', 'L', 'O', 'd', 'h', 'l', 'o', 'r', 'w']")
        self._run_test(
            "{a:Sort-reverse}",  # starts with 'S', contains 'r'
            "['w', 'r', 'o', 'l', 'h', 'd', 'O', 'L', 'L', 'E', ' ']")

    def test_specifier_arithmetic(self):
        self._run_test("{i:A+1}", "3")
        self._run_test("{i:A-1}", "1")
        self._run_test("{i:A*3}", "6")

    def test_specifier_conversions(self):
        self._run_test("{a:Cl}"   , "hello world")
        self._run_test("{h:CHC}"  , "Foo & Bar")
        self._run_test("{l:CSulc}", "A, b, c")

    def test_specifier_limit(self):
        self._run_test("{a:X20/ */}", "hElLo wOrLd")
        self._run_test("{a:X10/ */}", "hElLo wO *")

        with self.assertRaises(ValueError):
            self._run_test("{a:Xfoo/ */}", "hello wo *")

    def test_specifier_limit_bytes(self):
        self._run_test("{a:Xb20/ */}", "hElLo wOrLd")
        self._run_test("{a:Xb10/ */}", "hElLo wO *")

        self._run_test("{j:Xb50/〜/}", "げんそうきょう")
        self._run_test("{j:Xb20/〜/}", "げんそうき〜")
        self._run_test("{j:Xb20/ */}", "げんそうきょ *")

        with self.assertRaises(ValueError):
            self._run_test("{a:Xbfoo/ */}", "hello wo *")

    def test_specifier_map(self):
        self._run_test("{L:Mname/}" ,
                       "['John Doe', 'Jane Smith', 'Max Mustermann']")
        self._run_test("{L:Mage/}"  ,
                       "[42, 24, False]")

        self._run_test("{a:Mname}", self.kwdict["a"])
        self._run_test("{n:Mname}", "None")
        self._run_test("{title4:Mname}", "0")

        with self.assertRaises(ValueError):
            self._run_test("{t:Mname", "")

    def test_specifier_identity(self):
        self._run_test("{a:I}", self.kwdict["a"])
        self._run_test("{i:I}", self.kwdict["i"])
        self._run_test("{dt:I}", self.kwdict["dt"])

        self._run_test("{t!D:I}", self.kwdict["dt"])
        self._run_test("{t!D:I/O+01:30}", self.kwdict["dt"])
        self._run_test("{i:A+1/I}", self.kwdict["i"]+1)

    def test_chain_special(self):
        # multiple replacements
        self._run_test("{a:Rh/C/RE/e/RL/l/}", "Cello wOrld")
        self._run_test("{d[b]!s:R1/Q/R2/A/R0/Y/}", "Y")

        # join-and-replace
        self._run_test("{l:J-/Rb/E/}", "a-E-c")

        # join and slice
        self._run_test("{l:J-/[1:-1]}", "-b-")

        # optional-and-maxlen
        self._run_test("{d[a]:?</>/L1/too long/}", "<too long>")
        self._run_test("{d[c]:?</>/L5/too long/}", "")

        # parse and format datetime
        self._run_test("{ds:D%Y-%m-%dT%H:%M:%S%z/%Y%m%d}", "20100101")

        # sort and join
        self._run_test("{a:S/J}", " ELLOdhlorw")

        # map and join
        self._run_test("{L:Mname/J-}", "John Doe-Jane Smith-Max Mustermann")

    def test_separator(self):
        orig_separator = formatter._SEPARATOR
        try:
            formatter._SEPARATOR = "|"
            self._run_test("{a:Rh|C|RE|e|RL|l|}", "Cello wOrld")
            self._run_test("{d[b]!s:R1|Q|R2|A|R0|Y|}", "Y")

            formatter._SEPARATOR = "##"
            self._run_test("{l:J-##Rb##E##}", "a-E-c")
            self._run_test("{l:J-##[1:-1]}", "-b-")

            formatter._SEPARATOR = "\0"
            self._run_test("{d[a]:?<\0>\0L1\0too long\0}", "<too long>")
            self._run_test("{d[c]:?<\0>\0L5\0too long\0}", "")

            formatter._SEPARATOR = "?"
            self._run_test("{ds:D%Y-%m-%dT%H:%M:%S%z?%Y%m%d}", "20100101")
        finally:
            formatter._SEPARATOR = orig_separator

    def test_globals_env(self):
        os.environ["FORMATTER_TEST"] = value = self.kwdict["a"]

        self._run_test("{_env[FORMATTER_TEST]}"  , value)
        self._run_test("{_env[FORMATTER_TEST]!l}", value.lower())
        self._run_test("{z|_env[FORMATTER_TEST]}", value)

    def test_globals_now(self):
        fmt = formatter.parse("{_now}")
        out1 = fmt.format_map(self.kwdict)
        self.assertRegex(out1, r"^\d{4}-\d\d-\d\d \d\d:\d\d:\d\d(\.\d+)?$")

        out = formatter.parse("{_now:%Y%m%d}").format_map(self.kwdict)
        now = datetime.datetime.now()
        self.assertRegex(out, r"^\d{8}$")
        self.assertEqual(out, format(now, "%Y%m%d"))

        out = formatter.parse("{z|_now:%Y}").format_map(self.kwdict)
        self.assertRegex(out, r"^\d{4}$")
        self.assertEqual(out, format(now, "%Y"))

        out2 = fmt.format_map(self.kwdict)
        self.assertRegex(out1, r"^\d{4}-\d\d-\d\d \d\d:\d\d:\d\d(\.\d+)?$")
        self.assertNotEqual(out1, out2)

    def test_globals_nul(self):
        value = "None"

        self._run_test("{_nul}"         , value)
        self._run_test("{_nul[key]}"    , value)
        self._run_test("{z|_nul}"       , value)
        self._run_test("{z|_nul:%Y%m%s}", value)

    def test_literals(self):
        value = "foo"

        self._run_test("{'foo'}"      , value)
        self._run_test("{'foo'!u}"    , value.upper())
        self._run_test("{'f00':R0/o/}", value)

        self._run_test("{z|'foo'}"      , value)
        self._run_test("{z|''|'foo'}"   , value)
        self._run_test("{z|'foo'!u}"    , value.upper())
        self._run_test("{z|'f00':R0/o/}", value)

        self._run_test("{_lit[foo]}"       , value)
        self._run_test("{_lit[foo]!u}"     , value.upper())
        self._run_test("{_lit[f00]:R0/o/}" , value)
        self._run_test("{_lit[foobar][:3]}", value)
        self._run_test("{z|_lit[foo]}"     , value)

        # empty (#4492)
        self._run_test("{z|''}" , "")
        self._run_test("{''|''}", "")

        # special characters (dots, brackets, singlee quotes) (#5539)
        self._run_test("{'f.o.o'}"    , "f.o.o")
        self._run_test("{_lit[f.o.o]}", "f.o.o")
        self._run_test("{_lit[f'o'o]}", "f'o'o")
        self._run_test("{'f.[].[]'}"  , "f.[].[]")
        self._run_test("{z|'f.[].[]'}", "f.[].[]")

    def test_template(self):
        with tempfile.TemporaryDirectory() as tmpdirname:
            path1 = os.path.join(tmpdirname, "tpl1")
            path2 = os.path.join(tmpdirname, "tpl2")

            with open(path1, "w") as fp:
                fp.write("{a}")
            fmt1 = formatter.parse(f"\fT {path1}")

            with open(path2, "w") as fp:
                fp.write("{a!u:Rh/C/}\nFooBar")
            fmt2 = formatter.parse(f"\fT {path2}")

        self.assertEqual(fmt1.format_map(self.kwdict), self.kwdict["a"])
        self.assertEqual(fmt2.format_map(self.kwdict), "HELLO WORLD\nFooBar")

        with self.assertRaises(OSError):
            formatter.parse("\fT /")

    def test_expression(self):
        self._run_test("\fE a", self.kwdict["a"])
        self._run_test(
            "\fE name * 2 + ' ' + a",
            f"{self.kwdict['name']}{self.kwdict['name']} {self.kwdict['a']}")

    def test_fstring(self):
        self._run_test("\fF {a}", self.kwdict["a"])
        self._run_test(
            "\fF {name}{name} {a}",
            f"{self.kwdict['name']}{self.kwdict['name']} {self.kwdict['a']}")
        self._run_test(
            "\fF foo-'\"{a.upper()}\"'-bar",
            f"""foo-'"{self.kwdict['a'].upper()}"'-bar""")

    def test_template_fstring(self):
        with tempfile.TemporaryDirectory() as tmpdirname:
            path1 = os.path.join(tmpdirname, "tpl1")
            path2 = os.path.join(tmpdirname, "tpl2")

            with open(path1, "w") as fp:
                fp.write("{a}")
            fmt1 = formatter.parse(f"\fTF {path1}")

            with open(path2, "w") as fp:
                fp.write("foo-'\"{a.upper()}\"'-bar")
            fmt2 = formatter.parse(f"\fTF {path2}")

        self.assertEqual(fmt1.format_map(self.kwdict), self.kwdict["a"])
        self.assertEqual(fmt2.format_map(self.kwdict),
                         f"""foo-'"{self.kwdict['a'].upper()}"'-bar""")

        with self.assertRaises(OSError):
            formatter.parse("\fTF /")

    @unittest.skipIf(jinja2 is None, "no jinja2")
    def test_jinja(self):
        formatter.JinjaFormatter.env = None

        self._run_test("\fJ {{a}}", self.kwdict["a"])
        self._run_test(
            "\fJ {{name}}{{name}} {{a}}",
            f"{self.kwdict['name']}{self.kwdict['name']} {self.kwdict['a']}")
        self._run_test(
            "\fJ foo-'\"{{a | upper}}\"'-bar",
            f"""foo-'"{self.kwdict['a'].upper()}"'-bar""")

    @unittest.skipIf(jinja2 is None, "no jinja2")
    def test_template_jinja(self):
        formatter.JinjaFormatter.env = None

        with tempfile.TemporaryDirectory() as tmpdirname:
            path1 = os.path.join(tmpdirname, "tpl1")
            path2 = os.path.join(tmpdirname, "tpl2")

            with open(path1, "w") as fp:
                fp.write("{{a}}")
            fmt1 = formatter.parse(f"\fTJ {path1}")

            with open(path2, "w") as fp:
                fp.write("foo-'\"{{a | upper}}\"'-bar")
            fmt2 = formatter.parse(f"\fTJ {path2}")

        self.assertEqual(fmt1.format_map(self.kwdict), self.kwdict["a"])
        self.assertEqual(fmt2.format_map(self.kwdict),
                         f"""foo-'"{self.kwdict['a'].upper()}"'-bar""")

        with self.assertRaises(OSError):
            formatter.parse("\fTJ /")

    @unittest.skipIf(jinja2 is None, "no jinja2")
    def test_template_jinja_opts(self):
        formatter.JinjaFormatter.env = None

        with tempfile.TemporaryDirectory() as tmpdirname:
            path_filters = os.path.join(tmpdirname, "jinja_filters.py")
            path_template = os.path.join(tmpdirname, "jinja_template.txt")

            config.set((), "jinja", {
                "environment": {
                    "variable_start_string": "(((",
                    "variable_end_string"  : ")))",
                    "keep_trailing_newline": True,
                },
                "filters": path_filters,
            })

            with open(path_filters, "w") as fp:
                fp.write(r"""
import re

def datetime_format(value, format="%H:%M %d-%m-%y"):
    return value.strftime(format)

def sanitize(value):
    return re.sub(r"\s+", " ", value.strip())

__filters__ = {
    "dt_fmt": datetime_format,
    "sanitize_whitespace": sanitize,
}
""")

            with open(path_template, "w") as fp:
                fp.write("""\
Present Day  is ((( dt | dt_fmt("%B %d, %Y") )))
Present Time is ((( dt | dt_fmt("%H:%M:%S") )))

Hello ((( s | sanitize_whitespace ))).
I hope there is enough "(((S|sanitize_whitespace)))" for you.
""")
            fmt = formatter.parse(f"\fTJ {path_template}")

        self.assertEqual(fmt.format_map(self.kwdict), """\
Present Day  is January 01, 2010
Present Time is 00:00:00

Hello SPACE.
I hope there is enough "S P A C E" for you.
""")

    def test_module(self):
        with tempfile.TemporaryDirectory() as tmpdirname:
            path = os.path.join(tmpdirname, "testmod.py")

            with open(path, "w") as fp:
                fp.write("""
def gentext(kwdict):
    name = kwdict.get("Name") or kwdict.get("name") or "foo"
    return "'{title1}' by {}".format(name, **kwdict)

def lengths(kwdict):
    a = 0
    for k, v in kwdict.items():
        if k == k.lower():
            try:
                a += len(v)
            except TypeError:
                pass
    return format(a)

def noarg():
    return ""
""")
            sys.path.insert(0, tmpdirname)
            try:
                fmt1 = formatter.parse("\fM testmod:gentext")
                fmt2 = formatter.parse("\fM testmod:lengths")
                fmt0 = formatter.parse("\fM testmod:noarg")

                with self.assertRaises(AttributeError):
                    formatter.parse("\fM testmod:missing")
                with self.assertRaises(ImportError):
                    formatter.parse("\fM missing:missing")
            finally:
                sys.path.pop(0)

            fmt3 = formatter.parse(f"\fM {path}:gentext")
            fmt4 = formatter.parse(f"\fM {path}:lengths")

        self.assertEqual(fmt1.format_map(self.kwdict), "'Title' by Name")
        self.assertEqual(fmt2.format_map(self.kwdict), "139")

        self.assertEqual(fmt3.format_map(self.kwdict), "'Title' by Name")
        self.assertEqual(fmt4.format_map(self.kwdict), "139")

        with self.assertRaises(TypeError):
            self.assertEqual(fmt0.format_map(self.kwdict), "")

    def _run_test(self, format_string, result, default=None, fmt=format):
        fmt = formatter.parse(format_string, default, fmt)
        output = fmt.format_map(self.kwdict)
        self.assertEqual(output, result, format_string)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.

"""Extractors for https://www.bilibili.com/"""

from .common import Extractor, Message
from .. import text, util, exception


class BilibiliExtractor(Extractor):
    """Base class for bilibili extractors"""
    category = "bilibili"
    root = "https://www.bilibili.com"
    request_interval = (3.0, 6.0)

    def _init(self):
        self.api = BilibiliAPI(self)

    def items(self):
        for article in self.articles():
            article["_extractor"] = BilibiliArticleExtractor
            url = f"{self.root}/opus/{article['opus_id']}"
            yield Message.Queue, url, article

    def articles(self):
        return ()


class BilibiliArticleExtractor(BilibiliExtractor):
    """Extractor for a bilibili article"""
    subcategory = "article"
    pattern = (r"(?:https?://)?"
               r"(?:t\.bilibili\.com|(?:www\.)?bilibili.com/opus)/(\d+)")
    example = "https://www.bilibili.com/opus/12345"
    directory_fmt = ("{category}", "{username}")
    filename_fmt = "{id}_{num}.{extension}"
    archive_fmt = "{id}_{num}"

    def items(self):
        article_id = self.groups[0]
        article = self.api.article(article_id)

        # Flatten modules list
        modules = {}
        for module in article["detail"]["modules"]:
            if module["module_type"] == "MODULE_TYPE_BLOCKED":
                self.log.warning("%s: Blocked Article\n%s", article_id,
                                 module["module_blocked"].get("hint_message"))
            del module["module_type"]
            modules.update(module)
        article["detail"]["modules"] = modules

        article["username"] = modules["module_author"]["name"]

        pics = []

        if "module_top" in modules:
            try:
                pics.extend(modules["module_top"]["display"]["album"]["pics"])
            except Exception:
                pass

        if "module_content" in modules:
            for paragraph in modules["module_content"]["paragraphs"]:
                if "pic" not in paragraph:
                    continue

                try:
                    pics.extend(paragraph["pic"]["pics"])
                except Exception:
                    pass

        article["count"] = len(pics)
        yield Message.Directory, "", article

        livephoto = self.config("livephoto", True)
        for article["num"], pic in enumerate(pics, 1):
            url = pic["url"]
            article.update(pic)
            yield Message.Url, url, text.nameext_from_url(url, article)

            if livephoto and (url := pic.get("live_url")):
                article["id"] += "_l"
                yield Message.Url, url, text.nameext_from_url(url, article)


class BilibiliUserArticlesExtractor(BilibiliExtractor):
    """Extractor for a bilibili user's articles"""
    subcategory = "user-articles"
    pattern = (r"(?:https?://)?space\.bilibili\.com/(\d+)"
               r"/(?:article|upload/opus|dynamic)")
    example = "https://space.bilibili.com/12345/article"

    def articles(self):
        return self.api.user_articles(self.groups[0])


class BilibiliUserArticlesFavoriteExtractor(BilibiliExtractor):
    subcategory = "user-articles-favorite"
    pattern = (r"(?:https?://)?space\.bilibili\.com"
               r"/(\d+)/favlist\?fid=opus")
    example = "https://space.bilibili.com/12345/favlist?fid=opus"
    _warning = True

    def articles(self):
        if self._warning:
            if not self.cookies_check(("SESSDATA",)):
                self.log.error("'SESSDATA' cookie required")
            BilibiliUserArticlesFavoriteExtractor._warning = False
        return self.api.user_favlist()


class BilibiliAPI():
    def __init__(self, extractor):
        self.extractor = extractor

    def _call(self, endpoint, params):
        url = "https://api.bilibili.com/x/polymer/web-dynamic/v1" + endpoint
        data = self.extractor.request_json(url, params=params)

        if data["code"]:
            self.extractor.log.debug("Server response: %s", data)
            raise exception.AbortExtraction("API request failed")

        return data

    def user_articles(self, user_id):
        endpoint = "/opus/feed/space"
        params = {"host_mid": user_id}

        while True:
            data = self._call(endpoint, params)

            for item in data["data"]["items"]:
                params["offset"] = item["opus_id"]
                yield item

            if not data["data"]["has_more"]:
                break

    def article(self, article_id):
        url = "https://www.bilibili.com/opus/" + article_id

        while True:
            page = self.extractor.request(url).text
            try:
                return util.json_loads(text.extr(
                    page, "window.__INITIAL_STATE__=", "};") + "}")
            except Exception:
                if "window._riskdata_" not in page:
                    raise exception.AbortExtraction(
                        article_id + ": Unable to extract INITIAL_STATE data")
            self.extractor.wait(seconds=300)

    def user_favlist(self):
        endpoint = "/opus/feed/fav"
        params = {"page": 1, "page_size": 20}

        while True:
            data = self._call(endpoint, params)["data"]

            yield from data["items"]

            if not data.get("has_more"):
                break
            params["page"] += 1

    def login_user_id(self):
        url = "https://api.bilibili.com/x/space/v2/myinfo"
        data = self.extractor.request_json(url)

        if data["code"] != 0:
            self.extractor.log.debug("Server response: %s", data)
            raise exception.AbortExtraction(
                "API request failed. Are you logges in?")
        try:
            return data["data"]["profile"]["mid"]
        except Exception:
            raise exception.AbortExtraction("API request failed")



# -*- coding: utf-8 -*-

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.

"""Extractors for https://mangaread.org/"""

from .common import ChapterExtractor, MangaExtractor
from .. import text, exception


class MangareadBase():
    """Base class for Mangaread extractors"""
    category = "mangaread"
    root = "https://www.mangaread.org"

    def parse_chapter_string(self, chapter_string, data):
        match = text.re(
            r"(?:(.+)\s*-\s*)?[Cc]hapter\s*(\d+)(\.\d+)?(?:\s*-\s*(.+))?"
        ).match(text.unescape(chapter_string).strip())
        manga, chapter, minor, title = match.groups()
        manga = manga.strip() if manga else ""
        data["manga"] = data.pop("manga", manga)
        data["chapter"] = text.parse_int(chapter)
        data["chapter_minor"] = minor or ""
        data["title"] = title or ""
        data["lang"] = "en"
        data["language"] = "English"


class MangareadChapterExtractor(MangareadBase, ChapterExtractor):
    """Extractor for manga-chapters from mangaread.org"""
    pattern = (r"(?:https?://)?(?:www\.)?mangaread\.org"
               r"(/manga/[^/?#]+/[^/?#]+)")
    example = "https://www.mangaread.org/manga/MANGA/chapter-01/"

    def metadata(self, page):
        tags = text.extr(page, 'class="wp-manga-tags-list">', '</div>')
        data = {"tags": list(text.split_html(tags)[::2])}
        info = text.extr(page, '<h1 id="chapter-heading">', "</h1>")
        if not info:
            raise exception.NotFoundError("chapter")
        self.parse_chapter_string(info, data)
        return data

    def images(self, page):
        page = text.extr(
            page, '<div class="reading-content">', '<div class="entry-header')
        return [
            (text.extr(img, 'src="', '"').strip(), None)
            for img in text.extract_iter(page, '<img id="image-', '>')
        ]


class MangareadMangaExtractor(MangareadBase, MangaExtractor):
    """Extractor for manga from mangaread.org"""
    chapterclass = MangareadChapterExtractor
    pattern = r"(?:https?://)?(?:www\.)?mangaread\.org(/manga/[^/?#]+)/?$"
    example = "https://www.mangaread.org/manga/MANGA"

    def chapters(self, page):
        if 'class="error404' in page:
            raise exception.NotFoundError("manga")
        data = self.metadata(page)
        results = []
        for chapter in text.extract_iter(
                page, '<li class="wp-manga-chapter', "</li>"):
            url , pos = text.extract(chapter, '<a href="', '"')
            info, _ = text.extract(chapter, ">", "</a>", pos)
            self.parse_chapter_string(info, data)
            results.append((url, data.copy()))
        return results

    def metadata(self, page):
        extr = text.extract_from(text.extr(
            page, 'class="summary_content">', 'class="manga-action"'))
        return {
            "manga"      : text.extr(page, "<h1>", "</h1>").strip(),
            "description": text.unescape(text.remove_html(text.extract(
                page, ">", "</div>", page.index("summary__content"))[0])),
            "rating"     : text.parse_float(
                extr('total_votes">', "</span>").strip()),
            "manga_alt"  : text.remove_html(
                extr("Alternative\t\t</h5>\n\t</div>", "</div>")).split("; "),
            "author"     : list(text.extract_iter(
                extr('class="author-content">', "</div>"), '"tag">', "</a>")),
            "artist"     : list(text.extract_iter(
                extr('class="artist-content">', "</div>"), '"tag">', "</a>")),
            "genres"     : list(text.extract_iter(
                extr('class="genres-content">', "</div>"), '"tag">', "</a>")),
            "type"       : text.remove_html(
                extr("	Type	", "\n</div>")),
            "release"    : text.parse_int(text.remove_html(
                extr("	Release	", "\n</div>"))),
            "status"     : text.remove_html(
                extr("	Status	", "\n</div>")),
        }



# -*- coding: utf-8 -*-

# Copyright 2018-2026 Mike Fährmann
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.

"""Execute processes"""

from .common import PostProcessor
from .. import util, formatter
import subprocess
import os


if util.WINDOWS:
    def quote(s):
        s = s.replace('"', '\\"')
        return f'"{s}"'
else:
    from shlex import quote


def trim(args):
    return (args.partition(" ") if isinstance(args, str) else args)[0]


class ExecPP(PostProcessor):

    def __init__(self, job, options):
        PostProcessor.__init__(self, job)

        if cmds := options.get("commands"):
            self.cmds = [self._prepare_cmd(c) for c in cmds]
            execute = self.exec_many
        else:
            execute, self.args = self._prepare_cmd(options["command"])
            if options.get("async", False):
                self._exec = self._popen

        self.verbose = options.get("verbose", True)
        self.session = False
        self.creationflags = 0
        if options.get("session"):
            if util.WINDOWS:
                self.creationflags = subprocess.CREATE_NEW_PROCESS_GROUP
            else:
                self.session = True

        events = options.get("event")
        if events is None:
            events = ("after",)
        elif isinstance(events, str):
            events = events.split(",")
        job.register_hooks({event: execute for event in events}, options)

        if self._archive_init(job, options):
            self._archive_register(job)

    def _prepare_cmd(self, cmd):
        if isinstance(cmd, str):
            self._sub = util.re(
                r"(?i)\{(_directory(?:_unc)?|_filename"
                r"|_(?:temp)?path(?:_unc)?|)\}").sub
            return self.exec_string, cmd
        else:
            return self.exec_list, [formatter.parse(arg) for arg in cmd]

    def exec_list(self, pathfmt):
        archive = self.archive
        kwdict = pathfmt.kwdict

        if archive and archive.check(kwdict):
            return

        kwdict["_directory"] = pathfmt.directory
        kwdict["_filename"] = pathfmt.filename
        kwdict["_temppath"] = pathfmt.temppath
        kwdict["_path"] = pathfmt.path
        if util.WINDOWS:
            kwdict["_directory_unc"] = pathfmt.realdirectory
            kwdict["_path_unc"] = pathfmt.realpath

        args = [arg.format_map(kwdict) for arg in self.args]
        args[0] = os.path.expanduser(args[0])
        retcode = self._exec(args, False)

        if archive:
            archive.add(kwdict)
        return retcode

    def exec_string(self, pathfmt):
        archive = self.archive
        if archive and archive.check(pathfmt.kwdict):
            return

        self.pathfmt = pathfmt
        args = self._sub(self._replace, self.args)
        retcode = self._exec(args, True)

        if archive:
            archive.add(pathfmt.kwdict)
        return retcode

    def exec_many(self, pathfmt):
        if archive := self.archive:
            if archive.check(pathfmt.kwdict):
                return
            self.archive = False

        retcode = 0
        for execute, args in self.cmds:
            self.args = args
            if retcode := execute(pathfmt):
                # non-zero exit status
                break

        if archive:
            self.archive = archive
            archive.add(pathfmt.kwdict)
        return retcode

    def _exec(self, args, shell):
        if retcode := self._popen(args, shell).wait():
            self.log.warning("'%s' returned with non-zero exit status (%d)",
                             args if self.verbose else trim(args), retcode)
        return retcode

    def _popen(self, args, shell):
        self.log.debug("Running '%s'", args if self.verbose else trim(args))
        return util.Popen(
            args,
            shell=shell,
            creationflags=self.creationflags,
            start_new_session=self.session,
        )

    def _replace(self, match):
        attr = {
            ""              : "path",
            "_path"         : "path",
            "_path_unc"     : "realpath",
            "_temppath"     : "temppath",
            "_temppath_unc" : "temppath",
            "_directory"    : "directory",
            "_directory_unc": "realdirectory",
            "_filename"     : "filename",
        }[match[1].lower()]
        return quote(getattr(self.pathfmt, attr))


__postprocessor__ = ExecPP



# -*- coding: utf-8 -*-

# Copyright 2025 Mike Fährmann
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.

"""Extractors for https://ahottie.top/"""

from .common import Extractor, GalleryExtractor, Message
from .. import text

BASE_PATTERN = r"(?:https?://)?(?:www\.)?ahottie\.top"


class AhottieExtractor(Extractor):
    """Base class for ahottie extractors"""
    category = "ahottie"
    root = "https://ahottie.top"

    def items(self):
        for album in self.albums():
            yield Message.Queue, album["url"], album

    def _pagination(self, url, params):
        params["page"] = text.parse_int(params.get("page"), 1)

        while True:
            page = self.request(url, params=params).text

            for album in text.extract_iter(
                    page, '<div class="relative">', '</div>'):
                yield {
                    "url"  : text.extr(album, ' href="', '"'),
                    "title": text.unquote(text.extr(
                        album, ' alt="', '"')),
                    "date" : self.parse_datetime_iso(text.extr(
                        album, ' datetime="', '"')),
                    "_extractor": AhottieGalleryExtractor,
                }

            if 'rel="next"' not in page:
                break
            params["page"] += 1


class AhottieGalleryExtractor(GalleryExtractor, AhottieExtractor):
    directory_fmt = ("{category}", "{date:%Y-%m-%d} {title} ({gallery_id})")
    filename_fmt = "{num:>03}.{extension}"
    archive_fmt = "{gallery_id}_{num}_{filename}"
    pattern = BASE_PATTERN + r"(/albums/(\w+))"
    example = "https://ahottie.top/albums/1234567890"

    def metadata(self, page):
        extr = text.extract_from(page)
        return {
            "gallery_id": self.groups[1],
            "title": text.unescape(extr("<title>", "<").rpartition(" | ")[0]),
            "date" : self.parse_datetime_iso(extr('datetime="', '"')),
            "tags" : text.split_html(extr('<i ', '</div>'))[1:],
        }

    def images(self, page):
        data = {
            "_http_headers" : {"Referer": None},
            "_http_validate": self._validate,
        }

        results = []
        while True:
            pos = page.find("<time ") + 1
            for url in text.extract_iter(page, '" src="', '"', pos):
                results.append((url, data))

            pos = page.find('rel="next"', pos)
            if pos < 0:
                break
            page = self.request(text.rextr(page, 'href="', '"', pos)).text
        return results

    def _validate(self, response):
        hget = response.headers.get
        return not (
            hget("content-length") == "2421" and
            hget("content-type") == "image/jpeg"
        )


class AhottieTagExtractor(AhottieExtractor):
    subcategory = "tag"
    pattern = BASE_PATTERN + r"/tags/([^/?#]+)"
    example = "https://ahottie.top/tags/TAG"

    def albums(self):
        tag = self.groups[0]
        self.kwdict["search_tags"] = text.unquote(tag)
        return self._pagination(f"{self.root}/tags/{tag}", {})


class AhottieSearchExtractor(AhottieExtractor):
    subcategory = "search"
    pattern = BASE_PATTERN + r"/search/?\?([^#]+)"
    example = "https://ahottie.top/search?kw=QUERY"

    def albums(self):
        params = text.parse_query(self.groups[0])
        self.kwdict["search_tags"] = params.get("kw")
        return self._pagination(f"{self.root}/search", params)



# -*- coding: utf-8 -*-

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.

"""Extractors for https://thefap.net/"""

from .common import Extractor, Message
from .. import text, exception

BASE_PATTERN = r"(?:https?://)?(?:www\.)?thefap\.net"


class ThefapExtractor(Extractor):
    """Base class for thefap extractors"""
    category = "thefap"
    root = "https://thefap.net"
    directory_fmt = ("{category}", "{model_name} ({model_id})")
    filename_fmt = "{model}_{num:>03}.{extension}"
    archive_fmt = "{model_id}_{filename}"

    def _normalize_url(self, url):
        if not url:
            return ""
        url = url.strip()
        if "?w=" in url:
            url = url[:url.rfind("?")]
        elif url.endswith(":small"):
            url = url[:-6] + ":orig"
        if url.startswith("//"):
            url = "https:" + url
        elif url.startswith("/"):
            url = self.root + url
        return url


class ThefapPostExtractor(ThefapExtractor):
    """Extractor for individual thefap.net posts"""
    subcategory = "post"
    pattern = (BASE_PATTERN +
               r"(/([^/?#]+)-(\d+)/([^/?#]+)/i(\d+))")
    example = "https://thefap.net/MODEL-12345/KIND/i12345"

    def items(self):
        path, model, model_id, kind, post_id = self.groups

        page = self.request(self.root + path).text
        if "Not Found" in page:
            raise exception.NotFoundError("post")

        if model_name := text.extr(page, "<title>", " / "):
            model_name = text.unescape(model_name)
        else:
            model_name = text.unquote(model).replace(".", " ")

        data = {
            "model"     : model,
            "model_id"  : text.parse_int(model_id),
            "model_name": model_name,
            "kind"      : kind,
            "post_id"   : text.parse_int(post_id),
            "_http_headers": {"Referer": None},
        }
        yield Message.Directory, "", data

        data["num"] = 0
        page = text.extract(
            page, "\n</div>", "\n<!---->", page.index("</header>"))[0]
        for url in text.extract_iter(page, '<img src="', '"'):
            if url := self._normalize_url(url):
                data["num"] += 1
                yield Message.Url, url, text.nameext_from_url(url, data)


class ThefapModelExtractor(ThefapExtractor):
    """Extractor for thefap.net model pages"""
    subcategory = "model"
    pattern = BASE_PATTERN + r"/([^/?#]+)-(\d+)"
    example = "https://thefap.net/MODEL-12345/"

    def items(self):
        model, model_id = self.groups

        url = f"{self.root}/{model}-{model_id}/"
        page = self.request(url).text

        if 'id="content"' not in page:
            raise exception.NotFoundError("model")

        if model_name := text.extr(page, "<h2", "</h2>"):
            model_name = text.unescape(model_name[model_name.find(">")+1:])
        else:
            model_name = text.unquote(model).replace(".", " ")

        data = {
            "model"     : model,
            "model_id"  : text.parse_int(model_id),
            "model_name": model_name,
            "_http_headers": {"Referer": None},
        }
        yield Message.Directory, "", data

        base = f"{self.root}/ajax/model/{model_id}/page-"
        headers = {
            "X-Requested-With": "XMLHttpRequest",
            "Sec-Fetch-Dest"  : "empty",
            "Sec-Fetch-Mode"  : "cors",
            "Sec-Fetch-Site"  : "same-origin",
        }

        page = text.extr(page, '<div id="content"', '<div id="showmore"')
        imgs = text.extract_iter(page, 'data-src="', '"')
        pnum = 1
        data["num"] = 0

        while True:
            for url in imgs:
                if url := self._normalize_url(url):
                    data["num"] += 1
                    yield Message.Url, url, text.nameext_from_url(url, data)

            pnum += 1
            page = self.request(base + str(pnum), headers=headers).text
            if not page:
                break
            imgs = text.extract_iter(page, '<img src="', '"')



# -*- coding: utf-8 -*-

# Copyright 2026 Mike Fährmann
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.

"""Extractors for https://filester.me/"""

from .common import Extractor, Message
from .. import text
import random

BASE_PATTERN = r"(?:https?://)?(?:www\.)?filester\.me"


class FilesterExtractor(Extractor):
    """Base class for filester extractors"""
    category = "filester"
    archive_fmt = "{id}"
    root = "https://filester.me"

    def _download_url(self, slug):
        url = self.root + "/api/public/download"
        data = self.request_json(url, method="POST", json={"file_slug": slug})
        return (f"https://cache{random.choice((1, 6))}.filester.me"
                f"{data['download_url']}?download=true")


class FilesterFileExtractor(FilesterExtractor):
    subcategory = "file"
    filename_fmt = "{id} {filename}.{extension}"
    pattern = BASE_PATTERN + r"/d/([^/?#]+)"
    example = "https://filester.me/d/ID"

    def items(self):
        file_slug = self.groups[0]

        url = f"{self.root}/d/{file_slug}"
        page = self.request(url).text
        extr = text.extract_from(page)

        name = text.unquote(text.unescape(extr(
            'property="og:title" content="', '"')))
        file = text.nameext_from_name(name, {
            "id"  : file_slug,
            "uuid": extr('>UUID</span', '</span').rpartition(">")[2],
            "mime": extr('>Type</span', '</span').rpartition(">")[2],
            "date": self.parse_datetime_iso(extr(
                '>Uploaded</span', '</span').rpartition(">")[2]),
            "size": extr('>Size</span', '</span').rpartition(">")[2],
            "hash": extr('>SHA-256</span', '</span').rpartition(">")[2],
        })

        yield Message.Directory, "", file
        yield Message.Url, self._download_url(file_slug), file


class FilesterFolderExtractor(FilesterExtractor):
    subcategory = "folder"
    directory_fmt = ("{category}", "{folder_name} ({folder_id})")
    filename_fmt = "{num:>03} {id} {filename}.{extension}"
    pattern = BASE_PATTERN + r"/f/([^/?#]+)"
    example = "https://filester.me/f/ID"

    def items(self):
        folder_slug = self.groups[0]
        num = None

        url = f"{self.root}/f/{folder_slug}"
        params = {"page": 1}
        while True:
            page = self.request(url, params=params).text

            if num is None:
                extr = text.extract_from(page)
                kw = self.kwdict
                kw["folder_id"] = folder_slug
                kw["folder_name"] = text.unescape(extr(
                    'property="og:title" content="', '"'))
                kw["folder_uuid"] = extr('/t/', '"')
                kw["folder_date"] = self.parse_datetime(extr(
                    "<span>Created ", "<"), "%b %d, %Y")
                kw["count"] = num = text.parse_int(extr("<span>", " "))
                kw["folder_size"] = text.parse_bytes(extr("<span>", "B<"))
                del extr
                yield Message.Directory, "", {}

            for html in text.extract_iter(
                    page, 'class="file-item"', "</button>"):
                extr = text.extract_from(html)
                name = text.unescape(extr('data-name="', '"'))
                file = text.nameext_from_name(name, {
                    "size": extr('data-size="', '"'),
                    "date": self.parse_datetime_iso(extr('data-date="', '"')),
                    "id"  : extr("href='/d/", "'"),
                    "uuid": extr('src="/t/', '"'),
                    "num" : num,
                })
                num -= 1
                yield Message.Url, self._download_url(file["id"]), file

            if ">→</a>" not in page:
                break
            params["page"] += 1



# -*- coding: utf-8 -*-

# Copyright 2019-2025 Mike Fährmann
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.

"""Extractors for https://imgbb.com/"""

from .common import Extractor, Message
from .. import text, util, exception
from ..cache import cache


class ImgbbExtractor(Extractor):
    """Base class for imgbb extractors"""
    category = "imgbb"
    directory_fmt = ("{category}", "{user[name]:?//}{user[id]:? (/)/}",
                     "{album[title]} ({album[id]})")
    filename_fmt = "{title} ({id}).{extension}"
    archive_fmt = "{user[id]} {id}"
    cookies_domain = ".imgbb.com"
    cookies_names = ("PHPSESSID", "LID")
    root = "https://ibb.co"

    def items(self):
        self.login()

        for image in self.posts():
            url = image["url"]
            text.nameext_from_url(url, image)
            yield Message.Directory, "", image
            yield Message.Url, url, image

    def login(self):
        if self.cookies_check(self.cookies_names):
            return

        username, password = self._get_auth_info()
        if username:
            return self.cookies_update(self._login_impl(username, password))

    @cache(maxage=365*86400, keyarg=1)
    def _login_impl(self, username, password):
        self.log.info("Logging in as %s", username)

        url = "https://imgbb.com/login"
        page = self.request(url).text
        token = text.extr(page, 'name="auth_token" value="', '"')

        headers = {
            "Referer": url,
        }
        data = {
            "auth_token"   : token,
            "login-subject": username,
            "password"     : password,
        }
        response = self.request(url, method="POST", headers=headers, data=data)

        if not response.history:
            raise exception.AuthenticationError()
        return self.cookies

    def _pagination(self, page, url, params):
        seek, pos = text.extract(page, 'data-seek="', '"')
        tokn, pos = text.extract(page, 'PF.obj.config.auth_token="', '"', pos)
        resc, pos = text.extract(page, "CHV.obj.resource=", "};", pos)
        self.kwdict["user"] = util.json_loads(resc + "}").get("user")

        data = None
        while True:
            for obj in text.extract_iter(page, "data-object='", "'"):
                post = util.json_loads(text.unquote(obj))
                image = post["image"]
                image["filename"], image["name"] = \
                    image["name"], image["filename"]
                image["id"] = post["id_encoded"]
                image["title"] = post["title"]
                image["width"] = text.parse_int(post["width"])
                image["height"] = text.parse_int(post["height"])
                image["size"] = text.parse_int(image["size"])
                yield image

            if data:
                if not data["seekEnd"] or params["seek"] == data["seekEnd"]:
                    return
                params["seek"] = data["seekEnd"]
                params["page"] += 1
            elif not seek or 'class="pagination-next"' not in page:
                return
            else:
                params["action"] = "list"
                params["page"] = 2
                params["seek"] = seek
                params["auth_token"] = tokn

                headers = {
                    "Accept": "application/json, text/javascript, */*; q=0.01",
                    "X-Requested-With": "XMLHttpRequest",
                    "Origin": self.root,
                    "Sec-Fetch-Dest": "empty",
                    "Sec-Fetch-Mode": "cors",
                    "Sec-Fetch-Site": "same-origin",
                }

            data = self.request_json(
                url, method="POST", headers=headers, data=params)
            page = data["html"]


class ImgbbAlbumExtractor(ImgbbExtractor):
    """Extractor for imgbb albums"""
    subcategory = "album"
    pattern = r"(?:https?://)?ibb\.co/album/([^/?#]+)/?(?:\?([^#]+))?"
    example = "https://ibb.co/album/ID"

    def posts(self):
        album_id, qs = self.groups
        url = f"{self.root}/album/{album_id}"
        params = text.parse_query(qs)
        page = self.request(url, params=params).text
        extr = text.extract_from(page)

        self.kwdict["album"] = album = {
            "url": extr(
                'property="og:url" content="', '"'),
            "title": text.unescape(extr(
                'property="og:title" content="', '"')),
            "description": text.unescape(extr(
                'property="og:description" content="', '"')),
            "id": extr(
                'data-text="album-name" href="https://ibb.co/album/', '"'),
            "count": text.parse_int(extr(
                'data-text="image-count">', "<")),
        }

        url = self.root + "/json"
        params["pathname"] = "/album/" + album["id"]
        return self._pagination(page, url, params)


class ImgbbImageExtractor(ImgbbExtractor):
    subcategory = "image"
    pattern = r"(?:https?://)?ibb\.co/([^/?#]+)"
    example = "https://ibb.co/ID"

    def posts(self):
        url = f"{self.root}/{self.groups[0]}"
        page = self.request(url).text
        extr = text.extract_from(page)

        image = {
            "id"    : extr('property="og:url" content="https://ibb.co/', '"'),
            "title" : text.unescape(extr(
                '"og:title" content="', ' hosted at ImgBB"')),
            "url"   : extr('"og:image" content="', '"'),
            "width" : text.parse_int(extr('"og:image:width" content="', '"')),
            "height": text.parse_int(extr('"og:image:height" content="', '"')),
            "album" : extr("Added to <a", "</a>"),
            "date"  : self.parse_datetime_iso(extr('<span title="', '"')),
            "user"  : util.json_loads(extr(
                "CHV.obj.resource=", "};") + "}").get("user"),
        }

        if album := image["album"]:
            image["album"] = {
                "id"   : text.extr(album, "/album/", '"'),
                "title": text.unescape(album.rpartition(">")[2]),
            }
        else:
            image["album"] = None

        return (image,)


class ImgbbUserExtractor(ImgbbExtractor):
    """Extractor for imgbb user profiles"""
    subcategory = "user"
    directory_fmt = ("{category}", "{user[name]} ({user[id]})")
    pattern = r"(?:https?://)?([\w-]+)\.imgbb\.com/?(?:\?([^#]+))?"
    example = "https://USER.imgbb.com"

    def posts(self):
        user, qs = self.groups
        url = f"https://{user}.imgbb.com/"
        params = text.parse_query(qs)
        response = self.request(url, params=params, allow_redirects=False)

        if response.status_code < 300:
            params["pathname"] = "/"
            return self._pagination(response.text, url + "json", params)

        if response.status_code == 301:
            raise exception.NotFoundError("user")
        redirect = "HTTP redirect to " + response.headers.get("Location", "")
        if response.status_code == 302:
            raise exception.AuthRequired(
                ("username & password", "authenticated cookies"),
                "profile", redirect)
        raise exception.AbortExtraction(redirect)



# -*- coding: utf-8 -*-

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.

"""Extractors for https://fapachi.com/"""

from .common import Extractor, Message
from .. import text


class FapachiPostExtractor(Extractor):
    """Extractor for individual posts on fapachi.com"""
    category = "fapachi"
    subcategory = "post"
    root = "https://fapachi.com"
    directory_fmt = ("{category}", "{user}")
    filename_fmt = "{user}_{id}.{extension}"
    archive_fmt = "{user}_{id}"
    pattern = (r"(?:https?://)?(?:www\.)?fapachi\.com"
               r"/(?!search/)([^/?#]+)/media/(\d+)")
    example = "https://fapachi.com/MODEL/media/12345"

    def __init__(self, match):
        Extractor.__init__(self, match)
        self.user, self.id = match.groups()

    def items(self):
        data = {
            "user": self.user,
            "id"  : self.id,
        }
        page = self.request(f"{self.root}/{self.user}/media/{self.id}").text
        url = self.root + text.extract(
            page, 'data-src="', '"', page.index('class="media-img'))[0]
        yield Message.Directory, "", data
        yield Message.Url, url, text.nameext_from_url(url, data)


class FapachiUserExtractor(Extractor):
    """Extractor for all posts from a fapachi user"""
    category = "fapachi"
    subcategory = "user"
    root = "https://fapachi.com"
    pattern = (r"(?:https?://)?(?:www\.)?fapachi\.com"
               r"/(?!search(?:/|$))([^/?#]+)(?:/page/(\d+))?$")
    example = "https://fapachi.com/MODEL"

    def __init__(self, match):
        Extractor.__init__(self, match)
        self.user = match[1]
        self.num = text.parse_int(match[2], 1)

    def items(self):
        data = {"_extractor": FapachiPostExtractor}
        while True:
            url = f"{self.root}/{self.user}/page/{self.num}"
            page = self.request(url).text
            for post in text.extract_iter(page, 'model-media-prew">', ">"):
                if path := text.extr(post, '<a href="', '"'):
                    yield Message.Queue, self.root + path, data

            if '">Next page</a>' not in page:
                return
            self.num += 1



# -*- coding: utf-8 -*-

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.

"""Extractors for https://cfake.com/"""

from .common import Extractor, Message
from .. import text

BASE_PATTERN = r"(?:https?://)?(?:www\.)?cfake\.com"


class CfakeExtractor(Extractor):
    """Base class for cfake extractors"""
    category = "cfake"
    root = "https://cfake.com"
    directory_fmt = ("{category}", "{type}", "{type_name} ({type_id})")
    filename_fmt = "{category}_{type_name}_{id}.{extension}"
    archive_fmt = "{id}"

    def items(self):
        type, type_name, type_id, sub_id, pnum = self.groups

        if type.endswith("ies"):
            type = type[:-3] + "y"

        kwdict = self.kwdict
        kwdict["type"] = type
        kwdict["type_id"] = text.parse_int(type_id)
        kwdict["type_name"] = text.unquote(type_name).replace("_", " ")
        kwdict["sub_id"] = text.parse_int(sub_id)
        kwdict["page"] = pnum = text.parse_int(pnum, 1)
        yield Message.Directory, "", {}

        base = f"{self.root}/images/{type}/{type_name}/{type_id}"
        if sub_id:
            base = f"{base}/{sub_id}"

        while True:
            url = base if pnum < 2 else f"{base}/p{pnum}"
            page = self.request(url).text

            # Extract and yield images
            num = 0
            for image in self._extract_images(page):
                num += 1
                image["num"] = num + (pnum - 1) * 50
                url = image["url"]
                yield Message.Url, url, text.nameext_from_url(url, image)

            # Check for next page
            if not num or not (pnum := self._check_pagination(page)):
                return
            kwdict["page"] = pnum

    def _extract_images(self, page):
        """Extract image URLs and metadata from a gallery page"""
        for item in text.extract_iter(
                page, '<a href="javascript:showimage(', '</div></div>'):

            # Extract image path from showimage call
            # Format: 'big.php?show=2025/filename.jpg&id_picture=...
            show_param = text.extr(item, "show=", "&")
            if not show_param:
                continue

            # Extract metadata
            picture_id = text.extr(item, "id_picture=", "&")
            name_param = text.extr(item, "p_name=", "'")

            # Extract date
            date = text.extr(item, 'id="date_vignette">', '</div>')

            # Extract rating
            rating_text = text.extr(item, 'class="current-rating"', '</li>')
            rating = text.extr(rating_text, 'width:', 'px')

            # Convert thumbnail path to full image path
            # show_param is like "2025/filename.jpg"
            image_url = f"{self.root}/medias/photos/{show_param}"

            yield {
                "url": image_url,
                "id": text.parse_int(picture_id) if picture_id else 0,
                "name": text.unescape(name_param) if name_param else "",
                "date": date,
                "rating": rating,
            }

    def _check_pagination(self, page):
        """Check if there are more pages and return next page number"""
        # Look for current page indicator
        # Format: id="num_page_current" ><a href=".../ p1">1</a>
        current_section = text.extr(
            page, 'id="num_page_current"', '</div>')
        if not current_section:
            return None

        # Extract current page number from the link text
        current_page_str = text.extr(current_section, '">', '</a>')
        if not current_page_str:
            return None

        current_page = text.parse_int(current_page_str)
        if not current_page:
            return None

        next_page = current_page + 1

        # Check if next page link exists anywhere in the page
        # Look for href="/images/.../pN" pattern
        if f'/p{next_page}"' in page or f'/p{next_page} ' in page:
            return next_page

        return None


class CfakeCelebrityExtractor(CfakeExtractor):
    """Extractor for celebrity image galleries from cfake.com"""
    subcategory = "celebrity"
    pattern = (BASE_PATTERN + r"/images/(celebrity)"
               r"/([^/?#]+)/(\d+)()(?:/p(\d+))?")
    example = "https://cfake.com/images/celebrity/NAME/123"


class CfakeCategoryExtractor(CfakeExtractor):
    """Extractor for category image galleries from cfake.com"""
    subcategory = "category"
    pattern = (BASE_PATTERN + r"/images/(categories)"
               r"/([^/?#]+)/(\d+)()(?:/p(\d+))?")
    example = "https://cfake.com/images/categories/NAME/123"


class CfakeCreatedExtractor(CfakeExtractor):
    """Extractor for 'created' image galleries from cfake.com"""
    subcategory = "created"
    pattern = (BASE_PATTERN + r"/images/(created)"
               r"/([^/?#]+)/(\d+)/(\d+)(?:/p(\d+))?")
    example = "https://cfake.com/images/created/NAME/12345/123"


class CfakeCountryExtractor(CfakeExtractor):
    """Extractor for country image galleries from cfake.com"""
    subcategory = "country"
    pattern = (BASE_PATTERN + r"/images/(country)"
               r"/([^/?#]+)/(\d+)/(\d+)(?:/p(\d+))?")
    example = "https://cfake.com/images/country/NAME/12345/123"



# -*- coding: utf-8 -*-

# Copyright 2014-2025 Mike Fährmann
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.

"""Extractors for https://gelbooru.com/"""

from .common import Extractor, Message
from . import gelbooru_v02
from .. import text, exception
import binascii

BASE_PATTERN = r"(?:https?://)?(?:www\.)?gelbooru\.com/(?:index\.php)?\?"


class GelbooruBase():
    """Base class for gelbooru extractors"""
    category = "gelbooru"
    basecategory = "booru"
    root = "https://gelbooru.com"
    offset = 0

    def _api_request(self, params, key="post", log=False):
        if "s" not in params:
            params["s"] = "post"

        params["api_key"] = self.api_key
        params["user_id"] = self.user_id

        url = self.root + "/index.php?page=dapi&q=index&json=1"
        try:
            data = self.request_json(url, params=params)
        except exception.HttpError as exc:
            if exc.status == 401:
                raise exception.AuthRequired(
                    "'api-key' & 'user-id'", "the API")
            raise

        if not key:
            return data

        try:
            posts = data[key]
        except KeyError:
            if log:
                self.log.error("Incomplete API response (missing '%s')", key)
                self.log.debug("%s", data)
            return []

        if not isinstance(posts, list):
            return (posts,)
        return posts

    def _pagination(self, params):
        params["pid"] = self.page_start
        params["limit"] = self.per_page
        limit = self.per_page // 2
        pid = False

        if "tags" in params:
            tags = params["tags"].split()
            op = "<"
            id = False

            for tag in tags:
                if tag.startswith("sort:"):
                    if tag == "sort:id:asc":
                        op = ">"
                    elif tag == "sort:id" or tag.startswith("sort:id:"):
                        op = "<"
                    else:
                        pid = True
                elif tag.startswith("id:"):
                    id = True

            if not pid:
                if id:
                    tag = "id:" + op
                    tags = [t for t in tags if not t.startswith(tag)]
                tags = f"{' '.join(tags)} id:{op}"

        while True:
            posts = self._api_request(params)

            yield from posts

            if len(posts) < limit:
                return

            if pid:
                params["pid"] += 1
            else:
                if "pid" in params:
                    del params["pid"]
                params["tags"] = tags + str(posts[-1]["id"])

    def _pagination_html(self, params):
        url = self.root + "/index.php"
        params["pid"] = self.offset

        data = {}
        while True:
            num_ids = 0
            page = self.request(url, params=params).text

            for data["id"] in text.extract_iter(page, '" id="p', '"'):
                num_ids += 1
                yield from self._api_request(data)

            if num_ids < self.per_page:
                return
            params["pid"] += self.per_page

    def _file_url(self, post):
        url = post["file_url"]
        if url.endswith((".webm", ".mp4")):
            post["_fallback"] = (url,)
            md5 = post["md5"]
            root = text.root_from_url(post["preview_url"])
            path = f"/images/{md5[0:2]}/{md5[2:4]}/{md5}.webm"
            url = root + path
        return url

    def _notes(self, post, page):
        notes_data = text.extr(page, '<section id="notes"', '</section>')
        if not notes_data:
            return

        post["notes"] = notes = []
        extr = text.extract
        for note in text.extract_iter(notes_data, '<article', '</article>'):
            notes.append({
                "width" : int(extr(note, 'data-width="', '"')[0]),
                "height": int(extr(note, 'data-height="', '"')[0]),
                "x"     : int(extr(note, 'data-x="', '"')[0]),
                "y"     : int(extr(note, 'data-y="', '"')[0]),
                "body"  : extr(note, 'data-body="', '"')[0],
            })

    def _skip_offset(self, num):
        self.offset += num
        return num


class GelbooruTagExtractor(GelbooruBase,
                           gelbooru_v02.GelbooruV02TagExtractor):
    """Extractor for images from gelbooru.com based on search-tags"""
    pattern = BASE_PATTERN + r"page=post&s=list&tags=([^&#]*)"
    example = "https://gelbooru.com/index.php?page=post&s=list&tags=TAG"


class GelbooruPoolExtractor(GelbooruBase,
                            gelbooru_v02.GelbooruV02PoolExtractor):
    """Extractor for gelbooru pools"""
    per_page = 45
    pattern = BASE_PATTERN + r"page=pool&s=show&id=(\d+)"
    example = "https://gelbooru.com/index.php?page=pool&s=show&id=12345"

    skip = GelbooruBase._skip_offset

    def metadata(self):
        url = self.root + "/index.php"
        self._params = {
            "page": "pool",
            "s"   : "show",
            "id"  : self.pool_id,
        }
        page = self.request(url, params=self._params).text

        name, pos = text.extract(page, "<h3>Now Viewing: ", "</h3>")
        if not name:
            raise exception.NotFoundError("pool")

        return {
            "pool": text.parse_int(self.pool_id),
            "pool_name": text.unescape(name),
        }

    def posts(self):
        return self._pagination_html(self._params)


class GelbooruFavoriteExtractor(GelbooruBase,
                                gelbooru_v02.GelbooruV02FavoriteExtractor):
    """Extractor for gelbooru favorites"""
    per_page = 100
    pattern = BASE_PATTERN + r"page=favorites&s=view&id=(\d+)"
    example = "https://gelbooru.com/index.php?page=favorites&s=view&id=12345"

    skip = GelbooruBase._skip_offset

    def posts(self):
        # get number of favorites
        params = {
            "s"    : "favorite",
            "id"   : self.favorite_id,
            "limit": "2",
        }
        data = self._api_request(params, None, True)

        count = data["@attributes"]["count"]
        self.log.debug("API reports %s favorite entries", count)

        favs = data["favorite"]
        try:
            order = 1 if favs[0]["id"] < favs[1]["id"] else -1
        except LookupError as exc:
            self.log.debug(
                "Error when determining API favorite order (%s: %s)",
                exc.__class__.__name__, exc)
            order = -1
        else:
            self.log.debug("API yields favorites in %sscending order",
                           "a" if order > 0 else "de")

        order_favs = self.config("order-posts")
        if order_favs and order_favs[0] in ("r", "a"):
            self.log.debug("Returning them in reverse")
            order = -order

        if order < 0:
            return self._pagination(params, count)
        return self._pagination_reverse(params, count)

    def _pagination(self, params, count):
        if self.offset:
            pnum, skip = divmod(self.offset, self.per_page)
        else:
            pnum = skip = 0

        params["pid"] = pnum
        params["limit"] = self.per_page

        while True:
            favs = self._api_request(params, "favorite")

            if not favs:
                return

            if skip:
                favs = favs[skip:]
                skip = 0

            for fav in favs:
                for post in self._api_request({"id": fav["favorite"]}):
                    post["date_favorited"] = self.parse_timestamp(fav["added"])
                    yield post

            params["pid"] += 1

    def _pagination_reverse(self, params, count):
        pnum, last = divmod(count-1, self.per_page)
        if self.offset > last:
            # page number change
            self.offset -= last
            diff, self.offset = divmod(self.offset-1, self.per_page)
            pnum -= diff + 1
        skip = self.offset

        params["pid"] = pnum
        params["limit"] = self.per_page

        while True:
            favs = self._api_request(params, "favorite")
            favs.reverse()

            if skip:
                favs = favs[skip:]
                skip = 0

            for fav in favs:
                for post in self._api_request({"id": fav["favorite"]}):
                    post["date_favorited"] = self.parse_timestamp(fav["added"])
                    yield post

            params["pid"] -= 1
            if params["pid"] < 0:
                return


class GelbooruPostExtractor(GelbooruBase,
                            gelbooru_v02.GelbooruV02PostExtractor):
    """Extractor for single images from gelbooru.com"""
    pattern = (BASE_PATTERN +
               r"(?=(?:[^#]+&)?page=post(?:&|#|$))"
               r"(?=(?:[^#]+&)?s=view(?:&|#|$))"
               r"(?:[^#]+&)?id=(\d+)")
    example = "https://gelbooru.com/index.php?page=post&s=view&id=12345"


class GelbooruRedirectExtractor(GelbooruBase, Extractor):
    subcategory = "redirect"
    pattern = (r"(?:https?://)?(?:www\.)?gelbooru\.com"
               r"/redirect\.php\?s=([^&#]+)")
    example = "https://gelbooru.com/redirect.php?s=BASE64"

    def __init__(self, match):
        Extractor.__init__(self, match)
        self.url_base64 = match[1]

    def items(self):
        url = text.ensure_http_scheme(binascii.a2b_base64(
            self.url_base64).decode())
        data = {"_extractor": GelbooruPostExtractor}
        yield Message.Queue, url, data



# -*- coding: utf-8 -*-

# Copyright 2018-2025 Mike Fährmann
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.

"""Extractors for https://mangadex.org/"""

from .common import Extractor, Message
from .. import text, util, exception
from ..cache import cache, memcache
from collections import defaultdict

BASE_PATTERN = r"(?:https?://)?(?:www\.)?mangadex\.(?:org|cc)"


class MangadexExtractor(Extractor):
    """Base class for mangadex extractors"""
    category = "mangadex"
    directory_fmt = (
        "{category}", "{manga}",
        "{volume:?v/ />02}c{chapter:>03}{chapter_minor}{title:?: //}")
    filename_fmt = (
        "{manga}_c{chapter:>03}{chapter_minor}_{page:>03}.{extension}")
    archive_fmt = "{chapter_id}_{page}"
    root = "https://mangadex.org"
    useragent = util.USERAGENT_GALLERYDL
    _cache = {}

    def _init(self):
        self.uuid = self.groups[0]
        self.api = MangadexAPI(self)

    def _items_chapters(self):
        for chapter in self.chapters():
            uuid = chapter["id"]
            data = self._transform(chapter)
            data["_extractor"] = MangadexChapterExtractor
            self._cache[uuid] = data
            yield Message.Queue, f"{self.root}/chapter/{uuid}", data

    def _items_manga(self):
        data = {"_extractor": MangadexMangaExtractor}
        for manga in self.manga():
            url = f"{self.root}/title/{manga['id']}"
            yield Message.Queue, url, data

    items = _items_chapters

    def _transform(self, chapter):
        relationships = defaultdict(list)
        for item in chapter["relationships"]:
            relationships[item["type"]].append(item)

        cattributes = chapter["attributes"]
        if lang := cattributes.get("translatedLanguage"):
            lang = lang.partition("-")[0]

        if cattributes["chapter"]:
            chnum, sep, minor = cattributes["chapter"].partition(".")
        else:
            chnum, sep, minor = 0, "", ""

        return {
            **_manga_info(self, relationships["manga"][0]["id"]),
            "title"   : cattributes["title"],
            "volume"  : text.parse_int(cattributes["volume"]),
            "chapter" : text.parse_int(chnum),
            "chapter_minor": sep + minor,
            "chapter_id": chapter["id"],
            "date"    : self.parse_datetime_iso(cattributes["publishAt"]),
            "group"   : [group["attributes"]["name"]
                         for group in relationships["scanlation_group"]],
            "lang"    : lang,
            "count"   : cattributes["pages"],
            "_external_url": cattributes.get("externalUrl"),
        }


class MangadexCoversExtractor(MangadexExtractor):
    """Extractor for mangadex manga covers"""
    subcategory = "covers"
    directory_fmt = ("{category}", "{manga}", "Covers")
    filename_fmt = "{volume:>02}_{lang}.{extension}"
    archive_fmt = "c_{cover_id}"
    pattern = (BASE_PATTERN + r"/(?:title|manga)/(?!follows|feed$)([0-9a-f-]+)"
               r"(?:/[^/?#]+)?\?tab=art")
    example = ("https://mangadex.org/title"
               "/01234567-89ab-cdef-0123-456789abcdef?tab=art")

    def items(self):
        base = f"{self.root}/covers/{self.uuid}/"
        for cover in self.api.covers_manga(self.uuid):
            data = self._transform_cover(cover)
            name = data["cover"]
            text.nameext_from_url(name, data)
            data["cover_id"] = data["filename"]
            yield Message.Directory, "", data
            yield Message.Url, base + name, data

    def _transform_cover(self, cover):
        relationships = defaultdict(list)
        for item in cover["relationships"]:
            relationships[item["type"]].append(item)
        cattributes = cover["attributes"]

        return {
            **_manga_info(self, relationships["manga"][0]["id"]),
            "cover"   : cattributes["fileName"],
            "lang"    : cattributes.get("locale"),
            "volume"  : text.parse_int(cattributes["volume"]),
            "date"    : self.parse_datetime_iso(cattributes["createdAt"]),
            "date_updated": self.parse_datetime_iso(cattributes["updatedAt"]),
        }


class MangadexChapterExtractor(MangadexExtractor):
    """Extractor for manga-chapters from mangadex.org"""
    subcategory = "chapter"
    pattern = BASE_PATTERN + r"/chapter/([0-9a-f-]+)"
    example = ("https://mangadex.org/chapter"
               "/01234567-89ab-cdef-0123-456789abcdef")

    def items(self):
        try:
            data = self._cache.pop(self.uuid)
        except KeyError:
            chapter = self.api.chapter(self.uuid)
            data = self._transform(chapter)

        if data.get("_external_url") and not data["count"]:
            raise exception.AbortExtraction(
                f"Chapter {data['chapter']}{data['chapter_minor']} is not "
                f"available on MangaDex and can instead be read on the "
                f"official publisher's website at {data['_external_url']}.")

        yield Message.Directory, "", data

        if self.config("data-saver", False):
            path = "data-saver"
            key = "dataSaver"
        else:
            path = key = "data"

        server = self.api.athome_server(self.uuid)
        chapter = server["chapter"]
        base = f"{server['baseUrl']}/{path}/{chapter['hash']}/"

        enum = util.enumerate_reversed if self.config(
            "page-reverse") else enumerate
        for data["page"], path in enum(chapter[key], 1):
            text.nameext_from_url(path, data)
            yield Message.Url, base + path, data


class MangadexMangaExtractor(MangadexExtractor):
    """Extractor for manga from mangadex.org"""
    subcategory = "manga"
    pattern = BASE_PATTERN + r"/(?:title|manga)/(?!follows|feed$)([0-9a-f-]+)"
    example = ("https://mangadex.org/title"
               "/01234567-89ab-cdef-0123-456789abcdef")

    def items(self):
        items = self._items_chapters()
        if self.config("covers", False):
            import itertools
            url = f"{self.root}/title/{self.uuid}?tab=art"
            data = {"_extractor": MangadexCoversExtractor}
            items = itertools.chain(((Message.Queue, url, data),), items)
        return items

    def chapters(self):
        return self.api.manga_feed(self.uuid)


class MangadexFeedExtractor(MangadexExtractor):
    """Extractor for chapters from your Updates Feed"""
    subcategory = "feed"
    pattern = BASE_PATTERN + r"/titles?/feed$()"
    example = "https://mangadex.org/title/feed"

    def chapters(self):
        return self.api.user_follows_manga_feed()


class MangadexFollowingExtractor(MangadexExtractor):
    """Extractor for followed manga from your Library"""
    subcategory = "following"
    pattern = BASE_PATTERN + r"/titles?/follows(?:\?([^#]+))?$"
    example = "https://mangadex.org/title/follows"

    items = MangadexExtractor._items_manga

    def manga(self):
        return self.api.user_follows_manga()


class MangadexListExtractor(MangadexExtractor):
    """Extractor for mangadex MDLists"""
    subcategory = "list"
    pattern = (BASE_PATTERN +
               r"/list/([0-9a-f-]+)(?:/[^/?#]*)?(?:\?tab=(\w+))?")
    example = ("https://mangadex.org/list"
               "/01234567-89ab-cdef-0123-456789abcdef/NAME")

    def __init__(self, match):
        if match[2] == "feed":
            self.subcategory = "list-feed"
        else:
            self.items = self._items_manga
        MangadexExtractor.__init__(self, match)

    def chapters(self):
        return self.api.list_feed(self.uuid)

    def manga(self):
        return [
            item
            for item in self.api.list(self.uuid)["relationships"]
            if item["type"] == "manga"
        ]


class MangadexAuthorExtractor(MangadexExtractor):
    """Extractor for mangadex authors"""
    subcategory = "author"
    pattern = BASE_PATTERN + r"/author/([0-9a-f-]+)"
    example = ("https://mangadex.org/author"
               "/01234567-89ab-cdef-0123-456789abcdef/NAME")

    def items(self):
        for manga in self.api.manga_author(self.uuid):
            manga["_extractor"] = MangadexMangaExtractor
            url = f"{self.root}/title/{manga['id']}"
            yield Message.Queue, url, manga


class MangadexAPI():
    """Interface for the MangaDex API v5

    https://api.mangadex.org/docs/
    """

    def __init__(self, extr):
        self.extractor = extr
        self.headers = None
        self.headers_auth = {}

        self.username, self.password = extr._get_auth_info()
        if self.username:
            self.client_id = cid = extr.config("client-id")
            self.client_secret = extr.config("client-secret")
            if cid:
                self._authenticate_impl = self._authenticate_impl_client
            else:
                self._authenticate_impl = self._authenticate_impl_legacy
        else:
            self.authenticate = util.noop

        server = extr.config("api-server")
        self.root = ("https://api.mangadex.org" if server is None
                     else text.ensure_http_scheme(server).rstrip("/"))

    def athome_server(self, uuid):
        return self._call("/at-home/server/" + uuid)

    def author(self, uuid, manga=False):
        params = {"includes[]": ("manga",)} if manga else None
        return self._call("/author/" + uuid, params)["data"]

    def chapter(self, uuid):
        params = {"includes[]": ("scanlation_group",)}
        return self._call("/chapter/" + uuid, params)["data"]

    def covers_manga(self, uuid):
        params = {"manga[]": uuid}
        return self._pagination_covers("/cover", params)

    def list(self, uuid):
        return self._call("/list/" + uuid, None, True)["data"]

    def list_feed(self, uuid):
        return self._pagination_chapters(f"/list/{uuid}/feed", None, True)

    @memcache(keyarg=1)
    def manga(self, uuid):
        params = {"includes[]": ("artist", "author")}
        return self._call("/manga/" + uuid, params)["data"]

    def manga_author(self, uuid_author):
        params = {"authorOrArtist": uuid_author}
        return self._pagination_manga("/manga", params)

    def manga_feed(self, uuid):
        order = "desc" if self.extractor.config("chapter-reverse") else "asc"
        params = {
            "order[volume]" : order,
            "order[chapter]": order,
        }
        return self._pagination_chapters(f"/manga/{uuid}/feed", params)

    def user_follows_manga(self):
        params = {"contentRating": None}
        return self._pagination_manga(
            "/user/follows/manga", params, True)

    def user_follows_manga_feed(self):
        params = {"order[publishAt]": "desc"}
        return self._pagination_chapters(
            "/user/follows/manga/feed", params, True)

    def authenticate(self):
        self.headers_auth["Authorization"] = \
            self._authenticate_impl(self.username, self.password)

    @cache(maxage=900, keyarg=1)
    def _authenticate_impl_client(self, username, password):
        if refresh_token := _refresh_token_cache((username, "personal")):
            self.extractor.log.info("Refreshing access token")
            data = {
                "grant_type"   : "refresh_token",
                "refresh_token": refresh_token,
                "client_id"    : self.client_id,
                "client_secret": self.client_secret,
            }
        else:
            self.extractor.log.info("Logging in as %s", username)
            data = {
                "grant_type"   : "password",
                "username"     : self.username,
                "password"     : self.password,
                "client_id"    : self.client_id,
                "client_secret": self.client_secret,
            }

        self.extractor.log.debug("Using client-id '%s…'", self.client_id[:24])
        url = ("https://auth.mangadex.org/realms/mangadex"
               "/protocol/openid-connect/token")
        data = self.extractor.request_json(
            url, method="POST", data=data, fatal=None)

        try:
            access_token = data["access_token"]
        except Exception:
            raise exception.AuthenticationError(data.get("error_description"))

        if refresh_token != data.get("refresh_token"):
            _refresh_token_cache.update(
                (username, "personal"), data["refresh_token"])

        return "Bearer " + access_token

    @cache(maxage=900, keyarg=1)
    def _authenticate_impl_legacy(self, username, password):
        if refresh_token := _refresh_token_cache(username):
            self.extractor.log.info("Refreshing access token")
            url = self.root + "/auth/refresh"
            json = {"token": refresh_token}
        else:
            self.extractor.log.info("Logging in as %s", username)
            url = self.root + "/auth/login"
            json = {"username": username, "password": password}

        self.extractor.log.debug("Using legacy login method")
        data = self.extractor.request_json(
            url, method="POST", json=json, fatal=None)
        if data.get("result") != "ok":
            raise exception.AuthenticationError()

        if refresh_token != data["token"]["refresh"]:
            _refresh_token_cache.update(username, data["token"]["refresh"])
        return "Bearer " + data["token"]["session"]

    def _call(self, endpoint, params=None, auth=False):
        url = self.root + endpoint
        headers = self.headers_auth if auth else self.headers

        while True:
            if auth:
                self.authenticate()
            response = self.extractor.request(
                url, params=params, headers=headers, fatal=None)

            if response.status_code < 400:
                return response.json()
            if response.status_code == 429:
                until = response.headers.get("X-RateLimit-Retry-After")
                self.extractor.wait(until=until)
                continue

            msg = ", ".join(f'{error["title"]}: "{error["detail"]}"'
                            for error in response.json()["errors"])
            raise exception.AbortExtraction(
                f"{response.status_code} {response.reason} ({msg})")

    def _pagination_chapters(self, endpoint, params=None, auth=False):
        if params is None:
            params = {}

        lang = self.extractor.config("lang")
        if isinstance(lang, str) and "," in lang:
            lang = lang.split(",")
        params["translatedLanguage[]"] = lang
        params["includes[]"] = ("scanlation_group",)

        return self._pagination(endpoint, params, auth)

    def _pagination_manga(self, endpoint, params=None, auth=False):
        if params is None:
            params = {}

        return self._pagination(endpoint, params, auth)

    def _pagination_covers(self, endpoint, params=None, auth=False):
        if params is None:
            params = {}

        lang = self.extractor.config("lang")
        if isinstance(lang, str) and "," in lang:
            lang = lang.split(",")
        params["locales[]"] = lang
        params["contentRating"] = None
        params["order[volume]"] = \
            "desc" if self.extractor.config("chapter-reverse") else "asc"

        return self._pagination(endpoint, params, auth)

    def _pagination(self, endpoint, params, auth=False):
        config = self.extractor.config

        if "contentRating" not in params:
            ratings = config("ratings")
            if ratings is None:
                ratings = ("safe", "suggestive", "erotica", "pornographic")
            elif isinstance(ratings, str):
                ratings = ratings.split(",")
            params["contentRating[]"] = ratings
        params["offset"] = 0

        if api_params := config("api-parameters"):
            for key in api_params:
                if key in params:
                    del params[key]
            params.update(api_params)

        while True:
            data = self._call(endpoint, params, auth)
            yield from data["data"]

            params["offset"] = data["offset"] + data["limit"]
            if params["offset"] >= data["total"]:
                return


@cache(maxage=90*86400, keyarg=0)
def _refresh_token_cache(username):
    return None


@memcache(keyarg=1)
def _manga_info(self, uuid):
    manga = self.api.manga(uuid)

    rel = defaultdict(list)
    for item in manga["relationships"]:
        rel[item["type"]].append(item)
    mattr = manga["attributes"]

    return {
        "manga" : (mattr["title"].get("en") or
                   next(iter(mattr["title"].values()), "")),
        "manga_id": manga["id"],
        "manga_titles": [t.popitem()[1]
                         for t in mattr.get("altTitles") or ()],
        "manga_date"  : self.parse_datetime_iso(mattr.get("createdAt")),
        "description" : (mattr["description"].get("en") or
                         next(iter(mattr["description"].values()), "")),
        "demographic": mattr.get("publicationDemographic"),
        "origin": mattr.get("originalLanguage"),
        "status": mattr.get("status"),
        "year"  : mattr.get("year"),
        "rating": mattr.get("contentRating"),
        "links" : mattr.get("links"),
        "tags"  : [tag["attributes"]["name"]["en"] for tag in mattr["tags"]],
        "artist": [artist["attributes"]["name"] for artist in rel["artist"]],
        "author": [author["attributes"]["name"] for author in rel["author"]],
    }



# -*- coding: utf-8 -*-

# Copyright 2018-2025 Mike Fährmann
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.

"""Extractors for https://www.idolcomplex.com/"""

from . import sankaku

BASE_PATTERN = (r"(?:https?://)?(?:www\.)?"
                r"idol(?:\.sankaku)?complex\.com"
                r"(?:/[a-z]{2}(?:[-_][A-Z]{2})?)?")


class IdolcomplexBase():
    """Base class for idolcomplex extractors"""
    category = "idolcomplex"
    root = "https://www.idolcomplex.com"
    cookies_domain = ".idolcomplex.com"

    def _init(self):
        self.api = sankaku.SankakuAPI(self)
        self.api.ROOT = "https://i.sankakuapi.com"
        self.api.headers["Origin"] = self.root


class IdolcomplexTagExtractor(IdolcomplexBase, sankaku.SankakuTagExtractor):
    """Extractor for idolcomplex tag searches"""
    pattern = BASE_PATTERN + r"(?:/posts)?/?\?([^#]*)"
    example = "https://www.idolcomplex.com/en/posts?tags=TAGS"


class IdolcomplexPoolExtractor(IdolcomplexBase, sankaku.SankakuPoolExtractor):
    """Extractor for idolcomplex pools"""
    pattern = BASE_PATTERN + r"/pools?/(?:show/)?(\w+)"
    example = "https://www.idolcomplex.com/en/pools/0123456789abcdef"


class IdolcomplexPostExtractor(IdolcomplexBase, sankaku.SankakuPostExtractor):
    """Extractor for individual idolcomplex posts"""
    pattern = BASE_PATTERN + r"/posts?(?:/show)?/(\w+)"
    example = "https://www.idolcomplex.com/en/posts/0123456789abcdef"



# -*- coding: utf-8 -*-

# Copyright 2016-2023 Mike Fährmann
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.

"""Extractors for https://www.pixnet.net/"""

from .common import Extractor, Message
from .. import text, exception

BASE_PATTERN = r"(?:https?://)?(?!www\.)([\w-]+)\.pixnet.net"


class PixnetExtractor(Extractor):
    """Base class for pixnet extractors"""
    category = "pixnet"
    filename_fmt = "{num:>03}_{id}.{extension}"
    archive_fmt = "{id}"

    def __init__(self, match):
        Extractor.__init__(self, match)
        self.blog, self.item_id = self.groups
        self.root = f"https://{self.blog}.pixnet.net"

    def items(self):
        url = self.url_fmt.format(self.root, self.item_id)
        page = self.request(url, encoding="utf-8").text
        user = text.extr(page, '<meta name="author" content="', '";')
        data = {
            "blog": self.blog,
            "user": user.rpartition(" (")[0],
        }

        for info in self._pagination(page):
            url, pos = text.extract(info, ' href="', '"')
            alt, pos = text.extract(info, ' alt="', '"', pos)
            item = {
                "id"        : text.parse_int(url.rpartition("/")[2]),
                "title"     : text.unescape(alt),
                "_extractor": (PixnetFolderExtractor if "/folder/" in url else
                               PixnetSetExtractor),
            }
            item.update(data)
            yield Message.Queue, url, item

    def _pagination(self, page):
        while True:
            yield from text.extract_iter(page, '<li id="', '</li>')

            pnext = text.extr(page, 'class="nextBtn"', '>')
            if pnext is None and 'name="albumpass">' in page:
                raise exception.AbortExtraction(
                    f"Album {self.item_id} is password-protected.")
            if "href" not in pnext:
                return
            url = self.root + text.extr(pnext, 'href="', '"')
            page = self.request(url, encoding="utf-8").text


class PixnetImageExtractor(PixnetExtractor):
    """Extractor for a single photo from pixnet.net"""
    subcategory = "image"
    filename_fmt = "{id}.{extension}"
    directory_fmt = ("{category}", "{blog}")
    pattern = BASE_PATTERN + r"/album/photo/(\d+)"
    example = "https://USER.pixnet.net/album/photo/12345"

    def items(self):
        url = "https://api.pixnet.cc/oembed"
        params = {
            "url": (f"https://{self.blog}.pixnet.net"
                    f"/album/photo/{self.item_id}"),
            "format": "json",
        }

        data = self.request_json(url, params=params)
        data["id"] = text.parse_int(
            data["url"].rpartition("/")[2].partition("-")[0])
        data["filename"], _, data["extension"] = data["title"].rpartition(".")
        data["blog"] = self.blog
        data["user"] = data.pop("author_name")

        yield Message.Directory, "", data
        yield Message.Url, data["url"], data


class PixnetSetExtractor(PixnetExtractor):
    """Extractor for images from a pixnet set"""
    subcategory = "set"
    directory_fmt = ("{category}", "{blog}",
                     "{folder_id} {folder_title}", "{set_id} {set_title}")
    pattern = BASE_PATTERN + r"/album/set/(\d+)"
    example = "https://USER.pixnet.net/album/set/12345"

    def items(self):
        url = f"{self.root}/album/set/{self.item_id}"
        page = self.request(url, encoding="utf-8").text
        data = self.metadata(page)

        yield Message.Directory, "", data
        for num, info in enumerate(self._pagination(page), 1):
            url, pos = text.extract(info, ' href="', '"')
            src, pos = text.extract(info, ' src="', '"', pos)
            alt, pos = text.extract(info, ' alt="', '"', pos)

            photo = {
                "id": text.parse_int(url.rpartition("/")[2].partition("#")[0]),
                "url": src.replace("_s.", "."),
                "num": num,
                "filename": alt,
                "extension": src.rpartition(".")[2],
            }
            photo.update(data)
            yield Message.Url, photo["url"], photo

    def metadata(self, page):
        user , pos = text.extract(page, '<meta name="author" content="', '";')
        _    , pos = text.extract(page, 'id="breadcrumb"', '', pos)
        fid  , pos = text.extract(page, '/folder/', '"', pos)
        fname, pos = text.extract(page, '>', '<', pos)
        sid  , pos = text.extract(page, '/set/', '"', pos)
        sname, pos = text.extract(page, '>', '<', pos)
        return {
            "blog": self.blog,
            "user": user.rpartition(" (")[0],
            "folder_id"   : text.parse_int(fid, ""),
            "folder_title": text.unescape(fname).strip(),
            "set_id"      : text.parse_int(sid),
            "set_title"   : text.unescape(sname),
        }


class PixnetFolderExtractor(PixnetExtractor):
    """Extractor for all sets in a pixnet folder"""
    subcategory = "folder"
    url_fmt = "{}/album/folder/{}"
    pattern = BASE_PATTERN + r"/album/folder/(\d+)"
    example = "https://USER.pixnet.net/album/folder/12345"


class PixnetUserExtractor(PixnetExtractor):
    """Extractor for all sets and folders of a pixnet user"""
    subcategory = "user"
    url_fmt = "{}{}/album/list"
    pattern = BASE_PATTERN + r"()(?:/blog|/album(?:/list)?)?/?(?:$|[?#])"
    example = "https://USER.pixnet.net/"
